"""
Query scene release
"""

import re

from ... import utils

import logging  # isort:skip
_log = logging.getLogger(__name__)

natsort = utils.LazyModule(module='natsort', namespace=globals())


class SceneQuery:
    """
    Search query for scene release databases

    :param keywords: Search keywords
    :param group: Release group name
    :param episodes: :class:`~.release.Episodes`-like mapping

    .. note: `episodes` is only used to extract the actually wanted episodes
        from the search results. Season packs (e.g. "S03") is removed from
        `keywords`.

        PreDBs usually don't allow us to search for season packes or multiple
        episodes (or even multiple season packs), which we can do with this
        approach.
    """

    @classmethod
    def from_string(cls, string):
        """
        Create query from `string`

        :param string: Release name or path to release content

        `string` is passed to :class:`~.release.ReleaseInfo` and the result is
        passed to :meth:`from_release`.
        """
        release = utils.release.ReleaseInfo(string, strict=False)
        return cls.from_release(release)

    @classmethod
    def from_release(cls, release):
        """
        Create query from :class:`dict`-like object

        :param release: :class:`~.release.ReleaseName` or
            :class:`~.release.ReleaseInfo` instance or any :class:`dict`-like
            object with the keys ``type``, ``title``, ``year``, ``episodes``,
            ``resolution``, ``source``, ``video_codec`` and ``group``.
        """
        if isinstance(release, utils.release.ReleaseName):
            release = release.release_info

        original_release = release.copy()

        # Replace H.264/5 with H264/5
        release['video_codec'] = re.sub(r'\.', '', release.get('video_codec', ''))

        # Replace WEB-DL with WEB
        if release.get('source') == 'WEB-DL':
            release['source'] = 'WEB'

        # Group and episodes are handled separately from the other keywords. We
        # don't include episodes in the keywords. Instead, we get all episodes
        # of all seasons in one query and filter out the episodes we want when
        # postprocessing the results (see
        # PredbApiBase._postprocess_search_results()). This allows us to search
        # for multiple seasons ("S01S02") and multiple episodes
        # ("S01E01S02E02E03"). It also means we don't have to wait 40 seconds
        # per episode when querying srrdb for a season pack.
        needed_keys = utils.predbs.common.get_needed_keys(
            release,
            exclude=('group', 'episodes'),
        )

        keywords = [release[key] for key in needed_keys
                    if key in release]

        query = cls(
            *keywords,
            group=release.get('group', ''),
            episodes=release.get('episodes', {}),
            release_info=original_release,
        )
        return query

    def __init__(self, *keywords, group='', episodes={}, release_info=None):
        # Split each keyword at spaces
        kws = (k.strip()
               for kw in keywords
               for k in str(kw).split())

        # Remove season packs (keep "SxxEyy" but not "Sxx") because scene
        # releases are not season packs and we don't find anything if we look
        # for "S05". For season packs, we try to find all episodes of all
        # seasons and extract the relevant episodes in _handle_results().
        def exclude_season_pack(kw):
            if utils.release.Episodes.is_episodes_info(kw):
                episodes = utils.release.Episodes.from_string(kw)
                for season in tuple(episodes):
                    if not episodes[season]:
                        del episodes[season]
                return str(episodes)
            else:
                return kw

        kws = (exclude_season_pack(kw) for kw in kws)

        # Remove empty keywords
        # (I forgot why "-" is removed. Please explain if you know!)
        kws = (kw for kw in kws
               if kw and kw != '-')

        self._keywords = tuple(kws)
        self._group = str(group) if group else None
        self._episodes = episodes
        self._release_info = release_info

    @property
    def keywords(self):
        """Sequence of search terms"""
        return self._keywords

    @property
    def group(self):
        """Release group name"""
        return self._group

    @property
    def episodes(self):
        """:class:`~.release.Episodes`-like mapping"""
        return self._episodes

    @property
    def release_info(self):
        """:class:`~.ReleaseInfo` object this query was created from or `None`"""
        return self._release_info

    def __repr__(self):
        args = []
        if self.keywords:
            args.append(', '.join((repr(kw) for kw in self.keywords)))
        if self.group:
            args.append(f'group={self.group!r}')
        if self.episodes:
            args.append(f'episodes={self.episodes!r}')
        args_str = ', '.join(args)
        return f'{type(self).__name__}({args_str})'

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self.keywords == other.keywords
                and self.group == other.group
                and self.episodes == other.episodes
            )
        else:
            return NotImplemented



"""
Connect the user interface to the engine room
"""

from .base import JobBase, QueueJobBase  # isort:skip
from . import custom, dialog, imghost, mediainfo, poster, scene, screenshots, set, submit, torrent, webdb



"""
Release name parsing and formatting

:class:`ReleaseInfo` parses a string into a dictionary-like object with a
specific set of keys, e.g. "title", "resolution", "source", etc.

:class:`ReleaseName` wraps :class:`ReleaseInfo` to do the same, but in addition
it tries to read media data from the file system to get information. It also
adds a :class:`~.ReleaseName.format` method to turn everything back into a
string.
"""

import asyncio
import collections
import copy
import functools
import os
import re
import time

import natsort
import unidecode

from .. import constants, errors, utils
from . import LazyModule, fs, video
from .types import ReleaseType

import logging  # isort:skip
_log = logging.getLogger(__name__)

# Disable debugging messages from rebulk
logging.getLogger('rebulk').setLevel(logging.WARNING)

_guessit = LazyModule(module='guessit.api', name='_guessit', namespace=globals())


DELIM = r'[ \.-]'
"""
Regular expression that matches a single delimiter between release name
parts, usually ``"."`` or ``" "``
"""


class _translated_property:
    """
    Property that is automatically translated on lookup

    The translation is based on a mapping on the parent instance. It must be
    available as ``_translate``. It maps regular expressions to replacement
    strings. All regular expressions are applied in order.
    """

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        self.name = fget.__name__
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        else:
            value = self.fget(obj)
            return self._apply_translation(value, obj._translate)

    def _apply_translation(self, value, tables):
        def translate(string):
            table = tables.get(self.name, {})
            for regex, replacement in table.items():
                string = regex.sub(replacement, string)
            return string

        if isinstance(value, str):
            value = translate(value)
        elif isinstance(value, collections.abc.Iterable):
            value[:] = [translate(v) for v in value]

        return value

    def __set__(self, obj, value):
        if self.fset is None:
            raise AttributeError("Can't set attribute")
        self.fset(obj, value)

    def __delete__(self, obj):
        raise AttributeError("Can't delete attribute")

    def getter(self, fget):
        return type(self)(fget, self.fset, self.fdel, self.__doc__)

    def setter(self, fset):
        return type(self)(self.fget, fset, self.fdel, self.__doc__)


class ReleaseName(collections.abc.Mapping):
    """
    Standardized release name

    :param str path: Path to release file or directory

        If :attr:`path` exists, it is used to read video and audio metadata,
        e.g. to detect the codecs, resolution, etc.

    :param str name: Path or other string to pass to :class:`ReleaseInfo`
        (defaults to `path`)

    :param dict translate: Map names of properties that return a string
        (e.g. ``audio_format``) to maps of regular expressions to replacement
        strings. The replacement strings may contain backreferences to groups in
        their regular expression.

        Example:

        >>> {
        >>>     'audio_format': {
        >>>         re.compile(r'^DDP$'): r'DD+',
        >>>     },
        >>>     'video_format': {
        >>>         re.compile(r'^x26([45])$'): r'H.26\1',
        >>>     },
        >>> }

    :param str separator: Separator between release name parts (usually " " or
        ".") or `None` to use the default

    Example:

    >>> rn = ReleaseName("path/to/The.Foo.1984.1080p.Blu-ray.X264-ASDF")
    >>> rn.source
    'BluRay'
    >>> rn.format()
    'The Foo 1984 1080p BluRay DTS x264-ASDF'
    >>> "{title} ({year}) [{group}]".format(**rn)
    'The Foo (1984) [ASDF]'
    >>> rn.set_name('The Foo 1985 1080p BluRay DTS x264-AsdF')
    >>> "{title} ({year}) [{group}]".format(**rn)
    'The Foo (1985) [AsdF]'
    """

    def __init__(self, path, name=None, translate=None, separator=None):
        self._path = str(path)
        self._name = name
        self.separator = separator
        self._info = ReleaseInfo(str(name) if name is not None else self._path)
        self._translate = translate or {}

    @property
    def release_info(self):
        """Internal :class:`~.ReleaseInfo` instance"""
        return self._info

    def set_release_info(self, path):
        """
        Update internal :class:`ReleaseInfo` instance

        :param path: Argument for :class:`ReleaseInfo` (path or any other
            string)
        """
        self._info = ReleaseInfo(str(path))

    def __repr__(self):
        posargs = [self._path]
        kwargs = {}
        if self._name is not None:
            kwargs['name'] = self._name
        if self._translate:
            kwargs['translate'] = self._translate
        args = ', '.join(repr(arg) for arg in posargs)
        if kwargs:
            args += ', ' + ', '.join(f'{k}={v!r}' for k, v in kwargs.items())
        return f'{type(self).__name__}({args})'

    def __str__(self):
        return self.format()

    def __len__(self):
        return len(tuple(iter(self)))

    def __iter__(self):
        # Non-private properties are dictionary keys
        cls = type(self)
        return iter(attr for attr in dir(self)
                    if (not attr.startswith('_')
                        and isinstance(getattr(cls, attr), (property, _translated_property))))

    def __getitem__(self, name):
        if not isinstance(name, str):
            raise TypeError(f'Not a string: {name!r}')
        elif isinstance(getattr(type(self), name, None), (property, _translated_property)):
            return getattr(self, name)
        else:
            raise KeyError(name)

    @property
    def path(self):
        """:class:`str` version of the `path` instantiation argument"""
        return self._path

    @property
    def separator(self):
        """
        Separator between release name parts (usually " " or ".") or `None` to use
        the default
        """
        return self._separator

    @separator.setter
    def separator(self, separator):
        self._separator = str(separator) if separator is not None else ' '

    @property
    def type(self):
        """
        :class:`~.types.ReleaseType` enum or one of its value names

        See also :meth:`fetch_info`.
        """
        return self._info.get('type', ReleaseType.unknown)

    @type.setter
    def type(self, value):
        if not value:
            self._info['type'] = ReleaseType.unknown
        else:
            self._info['type'] = ReleaseType(value)

    @_translated_property
    def title(self):
        """
        Original name of movie or series or "UNKNOWN_TITLE"

        See also :meth:`fetch_info`.
        """
        return self._info.get('title') or 'UNKNOWN_TITLE'

    @title.setter
    def title(self, value):
        self._info['title'] = str(value)

    @_translated_property
    def title_aka(self):
        """
        Alternative name of movie or series or empty string

        For non-English original titles, this should be the English title. If
        :attr:`title` is identical, this is an empty string.

        See also :meth:`fetch_info`.
        """
        aka = self._info.get('aka') or ''
        if aka and aka != self.title:
            return aka
        else:
            return ''

    @title_aka.setter
    def title_aka(self, value):
        self._info['aka'] = str(value)

    @_translated_property
    def title_with_aka(self):
        """Combination of :attr:`title` and :attr:`title_aka`"""
        if self.title_aka:
            return f'{self.title} AKA {self.title_aka}'
        else:
            return self.title

    @_translated_property
    def title_with_aka_and_year(self):
        """
        Combination of :attr:`title_with_aka` with :attr:`country` and :attr:`year`
        if appropriate

        If :attr:`date` is specified, it is appended to :attr:`title_with_aka`.
        Otherwise, if :attr:`year_required` is `True`, :attr:`year` is appended.

        If :attr:`country_required` is `True`, :attr:`country` is appended.
        """
        title = [self.title_with_aka]

        if self.date:
            title.append(self.date)
        elif self.year_required:
            title.append(self.year)

        if self.country_required:
            title.append(self.country)

        return ' '.join(title)

    @_translated_property
    def year(self):
        """
        Release year or "UNKNOWN_YEAR" if :attr:`year_required` is set, empty string
        otherwise

        See also :meth:`fetch_info`.
        """
        if self.year_required:
            return self._info.get('year') or 'UNKNOWN_YEAR'
        else:
            return self._info.get('year') or ''

    @year.setter
    def year(self, value):
        if not isinstance(value, (str, int)) and value is not None:
            raise TypeError(f'Not a number: {value!r}')
        elif not value:
            self._info['year'] = ''
        else:
            year = str(value)
            current_year = int(time.strftime('%Y')) + 2
            if len(year) != 4 or not year.isdecimal() or not 1880 <= int(year) <= current_year:
                raise ValueError(f'Invalid year: {value}')
            self._info['year'] = year

    @property
    def year_required(self):
        """
        Whether :attr:`title_with_aka_and_year` includes :attr:`year`

        See also :meth:`fetch_info`.
        """
        default = self.type is ReleaseType.movie
        return getattr(self, '_year_required', default)

    @year_required.setter
    def year_required(self, value):
        self._year_required = bool(value)

    @_translated_property
    def country(self):
        """
        Release country or "UNKNOWN_COUNTRY" if :attr:`country_required` is set,
        empty string otherwise

        See also :meth:`fetch_info`.
        """
        country = self._info.get('country')
        if self.country_required:
            return country or 'UNKNOWN_COUNTRY'
        else:
            return country or ''

    @country.setter
    def country(self, value):
        if not value:
            self._info['country'] = ''
        else:
            self._info['country'] = str(value)

    @property
    def country_required(self):
        """
        Whether :attr:`title_with_aka_and_year` includes :attr:`country`

        See also :meth:`fetch_info`.
        """
        return getattr(self, '_country_required', False)

    @country_required.setter
    def country_required(self, value):
        self._country_required = bool(value)

    @_translated_property
    def episodes(self):
        """
        Season and episodes in "S01E02"-style format or "UNKNOWN_SEASON" for season
        packs, "UNKNOWN_EPISODE" for episodes, empty string for other types

        This property can be set to one or more season numbers (:class:`str`,
        :class:`int` or sequence of those), a "S01E02"-style string (see
        :meth:`Episodes.from_string`) or any falsy value.
        """
        if self.type is ReleaseType.season:
            return str(self._info.get('episodes') or 'UNKNOWN_SEASON')
        elif self.type is ReleaseType.episode:
            episodes = self._info.get('episodes', Episodes())
            if not any(season for season in episodes.values()):
                return 'UNKNOWN_EPISODE'
            else:
                return str(episodes)
        elif self.type is ReleaseType.unknown:
            return str(self._info.get('episodes') or '')
        else:
            return ''

    @episodes.setter
    def episodes(self, value):
        if isinstance(value, str) and Episodes.has_episodes_info(value):
            self._info['episodes'] = Episodes.from_string(value)
        elif not isinstance(value, str) and isinstance(value, collections.abc.Mapping):
            self._info['episodes'] = Episodes(value)
        elif not isinstance(value, str) and isinstance(value, collections.abc.Iterable):
            self._info['episodes'] = Episodes({v: () for v in value})
        elif value:
            self._info['episodes'] = Episodes({value: ()})
        else:
            self._info['episodes'] = Episodes()

    @property
    def episodes_dict(self):
        """Episodes as :class:`Episodes` object"""
        return self._info.get('episodes') or Episodes()

    @property
    def only_season(self):
        """
        Season if there is only one season in :attr:`episodes_dict`, `None`
        otherwise
        """
        seasons = tuple(self.episodes_dict)
        return seasons[0] if len(seasons) == 1 else None

    @_translated_property
    def episode_title(self):
        """Episode title if :attr:`type` is "episode" or empty string"""
        if self.type is ReleaseType.episode:
            return self._info.get('episode_title') or ''
        else:
            return ''

    @episode_title.setter
    def episode_title(self, value):
        self._info['episode_title'] = str(value)

    @_translated_property
    def date(self):
        """
        Date (YYYY-MM-DD)

        Sometimes single episodes are not released as part of a season but with
        an air date to identify them.

        For episodes, this should be the air date or an empty string.

        For anything that isn't an episode, this is always an empty string.
        """
        if self.type is ReleaseType.episode:
            return self._info.get('date') or ''
        else:
            return ''

    @date.setter
    def date(self, value):
        self._info['date'] = str(value)

    @_translated_property
    def service(self):
        """Streaming service abbreviation (e.g. "AMZN", "NF") or empty string"""
        return self._info.get('service') or ''

    @service.setter
    def service(self, value):
        self._info['service'] = str(value)

    @_translated_property
    def edition(self):
        """
        List of "Director's Cut", "Uncut", "Unrated", etc

        :raise ContentError: if :attr:`path` exists but contains unexpected data
        """
        if 'edition' not in self._info:
            self._info['edition'] = []

        # Dual Audio
        while 'Dual Audio' in self._info['edition']:
            self._info['edition'].remove('Dual Audio')
        if self.has_dual_audio:
            self._info['edition'].append('Dual Audio')

        return self._info['edition']

    @edition.setter
    def edition(self, value):
        self._info['edition'] = [str(v) for v in value]

    @_translated_property
    def source(self):
        '''Original source (e.g. "BluRay", "WEB-DL") or "UNKNOWN_SOURCE"'''
        source = None

        # DVD
        video_ts_path = utils.fs.find_name('VIDEO_TS', self._path, validator=os.path.isdir)
        if video_ts_path:
            # Include VIDEO_TS siblings (e.g. "cover")
            release_size = utils.fs.path_size(os.path.dirname(video_ts_path))
            _log.debug('VIDEO_TS size: %r: %r', video_ts_path, release_size)
            if release_size <= 4_700_000_000:
                source = 'DVD5'
            else:
                source = 'DVD9'

        if source is None:
            # Blu-ray
            bdmv_path = utils.fs.find_name('BDMV', self._path, validator=os.path.isdir)
            if bdmv_path:
                # Include BDMV siblings (e.g. "CERTIFICATE", and "ANY!")
                release_size = utils.fs.path_size(os.path.dirname(bdmv_path))
                _log.debug('Blu-ray size: %r: %r', bdmv_path, release_size)
                if release_size <= 25e9:
                    source = 'BD25'
                elif release_size <= 50e9:
                    source = 'BD50'
                elif release_size <= 66e9:
                    source = 'BD66'
                else:
                    source = 'BD100'

        if source is None:
            source = self._info.get('source') or 'UNKNOWN_SOURCE'

        return source

    @source.setter
    def source(self, value):
        self._info['source'] = str(value)

    @_translated_property
    def resolution(self):
        """
        Resolution (e.g. "1080p") or "UNKNOWN_RESOLUTION"

        :raise ContentError: if :attr:`path` exists but contains unexpected data
        """
        res = video.resolution(self._path, default=None)
        if res is None:
            res = self._info.get('resolution') or 'UNKNOWN_RESOLUTION'
        return res

    @resolution.setter
    def resolution(self, value):
        self._info['resolution'] = str(value)

    @_translated_property
    def audio_format(self):
        """
        Audio format (e.g. "FLAC") or empty string (no audio track)

        :raise ContentError: if :attr:`path` exists but contains unexpected data
        """
        af = video.audio_format(self._path, default=None)
        if af is None:
            af = self._info.get('audio_codec') or ''
        return af

    @audio_format.setter
    def audio_format(self, value):
        self._info['audio_codec'] = str(value)

    @_translated_property
    def audio_channels(self):
        """
        Audio channels (e.g. "5.1") or empty string (no audio track)

        :raise ContentError: if :attr:`path` exists but contains unexpected data
        """
        ac = video.audio_channels(self._path, default=None)
        if ac is None:
            ac = self._info.get('audio_channels') or ''
        return ac

    @audio_channels.setter
    def audio_channels(self, value):
        self._info['audio_channels'] = str(value)

    @_translated_property
    def hdr_format(self):
        """
        HDR format name (e.g. "Dolby Vision" or "HDR10") or empty string

        If not set explicitly and :attr:`path` exists, this value is
        autodetected if possible, otherwise default to whatever
        :class:`ReleaseInfo` detected in the release name.

        Setting this value back to `None` turns on autodetection as described
        above.

        :raise ContentError: if :attr:`path` exists but contains unexpected data
        """
        # Use manually set value unless it is None
        if getattr(self, '_hdr_format', None) is not None:
            return self._hdr_format

        # Autodetect from existing files
        if os.path.exists(self._path):
            hdr_format = video.hdr_format(self._path, default=None)
            if hdr_format:
                self._hdr_format = hdr_format
                return self._hdr_format

        # Autodetect from release name
        guessed_hdr = self._info.get('hdr_format', '')
        if guessed_hdr:
            return guessed_hdr

        # No HDR detected
        return ''

    @hdr_format.setter
    def hdr_format(self, value):
        if value is None:
            # Turn on autodetection
            self._hdr_format = None
        elif value == '':
            # No HDR
            self._hdr_format = ''
        elif value in video.hdr_formats:
            self._hdr_format = value
        else:
            raise ValueError(f'Unknown HDR format: {value!r}')

    @_translated_property
    def video_format(self):
        """
        Video format (or encoder in case of x264/x265/XviD) or "UNKNOWN_VIDEO_FORMAT"

        :raise ContentError: if :attr:`path` exists but contains unexpected data
        """
        vf = video.video_format(self._path, default=None)
        if vf is None:
            vf = self._info.get('video_codec') or 'UNKNOWN_VIDEO_FORMAT'
        return vf

    @video_format.setter
    def video_format(self, value):
        self._info['video_codec'] = str(value)

    @_translated_property
    def group(self):
        '''Name of release group or "NOGROUP"'''
        return self._info.get('group') or 'NOGROUP'

    @group.setter
    def group(self, value):
        self._info['group'] = str(value)

    @property
    def has_commentary(self):
        """
        Whether this release has a commentary audio track

        If not set explicitly and `:attr:`path`` exists, this value is
        autodetected by looking for "commentary" case-insensitively in any audio
        track title.

        If not set explicitly and :attr:`path` does not exists, default to
        detection by :class:`ReleaseInfo`.

        Setting this value back to `None` turns on autodetection as described
        above.

        :raise ContentError: if :attr:`path` exists but contains unexpected data
        """
        # Use manually set value unless it is None
        if getattr(self, '_has_commentary', None) is not None:
            return self._has_commentary

        # Find "Commentary" in audio track titles
        elif os.path.exists(self._path):
            self._has_commentary = bool(video.has_commentary(self._path))
            return self._has_commentary

        # Default to ReleaseInfo['has_commentary']
        else:
            return self._info.get('has_commentary')

    @has_commentary.setter
    def has_commentary(self, value):
        if value is None:
            self._has_commentary = None
        else:
            self._has_commentary = bool(value)

    @property
    def has_dual_audio(self):
        """
        Whether this release has an English and a non-English audio track

        If not set explicitly and :attr:`path` exists, this value is
        autodetected if possible, otherwise default to whatever
        :class:`ReleaseInfo` detected in the release name.

        Setting this value back to `None` turns on autodetection as described
        above.

        :raise ContentError: if :attr:`path` exists but contains unexpected data
        """
        # Use manually set value unless it is None
        if getattr(self, '_has_dual_audio', None) is not None:
            return self._has_dual_audio

        # Autodetect dual audio
        elif os.path.exists(self._path):
            self._has_dual_audio = bool(video.has_dual_audio(self._path))
            return self._has_dual_audio

        # Default to ReleaseInfo['edition']
        else:
            return 'Dual Audio' in self._info.get('edition', ())

    @has_dual_audio.setter
    def has_dual_audio(self, value):
        if value is None:
            self._has_dual_audio = None
        else:
            self._has_dual_audio = bool(value)

    _needed_attrs = {
        ReleaseType.movie: ('title', 'year', 'resolution', 'source', 'video_format'),
        ReleaseType.season: ('title', 'episodes', 'resolution', 'source', 'video_format'),
        ReleaseType.episode: ('title', 'episodes', 'resolution', 'source', 'video_format'),
    }

    @functools.cached_property
    def dvd_encoding(self):
        """"PAL", "NTSC" or `None`"""
        if self.source in ('DVD', 'DVD5', 'DVD9', 'DVDRip'):
            frame_rate = utils.video.frame_rate(self._path, default=0)
            if frame_rate:
                if frame_rate <= 25:
                    return 'PAL'
                elif frame_rate <= 30:
                    return 'NTSC'

    @property
    def is_complete(self):
        """
        Whether all needed information is known and the string returned by
        :meth:`format` will not contain "UNKNOWN_*"

        This always returns `False` if :attr:`type` is
        :attr:`~.ReleaseType.unknown`.
        """
        if self.type is ReleaseType.unknown:
            return False
        for attr in self._needed_attrs[self.type]:
            if self[attr].startswith('UNKNOWN_'):
                return False
        return True

    async def fetch_info(self, *, webdb, webdb_id, callback=None):
        """
        Fill in information web database

        :param webdb: :class:`~.webdbs.base.WebDbApiBase` instance (see
            :class:`~.webdbs.webdb`)
        :param str webdb_id: Valid ID for `webdb`
        :param callable callback: Function to call after info was fetched; gets
            the instance (`self`) as a keyword argument

        Attempt to set these attributes if they are supported by `webdb`:

          - :attr:`title`
          - :attr:`title_aka`
          - :attr:`type`
          - :attr:`year`
          - :attr:`year_required` (IMDb only)
          - :attr:`country`
          - :attr:`country_required` (IMDb only)

        :return: The method's instance (`self`) for convenience
        """
        await asyncio.gather(
            self._update_attributes(webdb, webdb_id),
            self._update_type(webdb, webdb_id),
            self._update_country(webdb, webdb_id),
            self._update_year_country_required(webdb, webdb_id),
        )
        _log.debug('Release name updated with %s info: %s', webdb.label, self)
        if callback is not None:
            callback(self)
        return self

    async def _update_attributes(self, webdb, webdb_id):
        info = await webdb.gather(
            webdb_id,
            'title_english',
            'title_original',
            'year',
        )
        for attr, key in (
            ('title', 'title_original'),
            ('title_aka', 'title_english'),
            ('year', 'year'),
        ):
            # Only overload non-empty values
            if info[key]:
                setattr(self, attr, info[key])

    async def _update_type(self, webdb, webdb_id):
        # Use type from database if possible. ReleaseInfo can misdetect (e.g. if
        # mini-series doesn't contain "S01"). But if ReleaseInfo detected an
        # episode (e.g. "S04E03"), it's unlikely to be wrong.
        webdb_type = await webdb.type(webdb_id)
        if webdb_type and self.type is not ReleaseType.episode:
            self.type = webdb_type

    async def _update_country(self, webdb, webdb_id):
        try:
            countries = await webdb.countries(webdb_id)
        except NotImplementedError:
            pass
        else:
            if countries:
                self.country = utils.country.tld(countries[0]).upper()

    async def _update_year_country_required(self, webdb, webdb_id):
        if self.type in (ReleaseType.season, ReleaseType.episode):
            # Find out if there are any TV show name duplicates. If so, require
            # year and/or country code in release name to pin it down.

            @functools.lru_cache(maxsize=None)
            def normalize_title(title):
                return unidecode.unidecode(title.casefold())

            # Find result with the same title, removing any "smart" matches
            query = utils.webdbs.Query(title=self.title, type=ReleaseType.series)
            results = [
                {
                    'title': normalize_title(result.title),
                    'year': result.year,
                    'countries': await result.countries(),
                }
                for result in await webdb.search(query)
                if normalize_title(self.title) == normalize_title(result.title)
            ]

            # Assemble relevant info
            info = {
                'title': normalize_title(self.title),
                'year': await webdb.year(webdb_id),
                'countries': await webdb.countries(webdb_id),
            }

            def has_duplicates(*attributes):
                combinations = [
                    combine(result, attributes)
                    for result in results
                ]
                relevant_info = {attr: info[attr] for attr in attributes}
                return combinations.count(relevant_info) >= 2

            def combine(result, attributes):
                combined = {}
                for attr in attributes:
                    if result.get(attr):
                        combined[attr] = result[attr]
                return combined

            if has_duplicates('title'):
                if not has_duplicates('title', 'year'):
                    self.year_required = True
                elif not has_duplicates('title', 'countries'):
                    self.country_required = True
                else:
                    self.year_required = True
                    self.country_required = True

    def format(self, separator=None):
        """
        Assemble all parts into string

        :param separator: Separator to use instead of space

            The typical use case would be ``.``.
        """
        parts = [self.title_with_aka_and_year]

        if self.type in (ReleaseType.season, ReleaseType.episode) and not self.date:
            parts.append(str(self.episodes))

        if self.edition:
            parts.append(' '.join(self.edition))

        if self.dvd_encoding:
            parts.append(self.dvd_encoding)
        else:
            parts.append(self.resolution)

        if self.service:
            parts.append(self.service)

        parts.append(self.source)

        if self.audio_format:
            parts.append(self.audio_format)

        if self.audio_channels:
            parts.append(self.audio_channels)

        if self.hdr_format:
            parts.append(self.hdr_format)

        parts.append(self.video_format)

        joined = ' '.join(parts) + f'-{self.group}'

        # Separator (usually " " or ".")
        sep = separator if separator is not None else self.separator
        if sep != ' ':
            joined = joined.replace(' ', sep)

        if sep == '.':
            # Replace "&" with "and" before "&" gets removed
            joined = re.sub(r'&', 'and', joined)
            # Remove non-word characters like (e.g.: ,:) with some exceptions
            joined = re.sub(r'[^ \w\.-]', '', joined)
            # Deduplicate "."
            joined = re.sub(r'\.+', '.', joined)

        return joined


class ReleaseInfo(collections.abc.MutableMapping):
    """
    Parse information from release name or path

    .. note::

       Consider using :class:`~.ReleaseName` instead to get more accurate info
       from the data of existing files.

    :param str release: Release name or path to release content

    :param bool strict: Whether to raise :class:`~.errors.ContentError` if
        `release` looks bad, e.g. an abbreviated scene release file name like
        "tf-foof.mkv"

    If `release` looks like an abbreviated scene file name (e.g.
    "abd-mother.mkv"), the parent directory's name is used if possible.

    Gathered information is provided as a dictionary with the following keys:

      - ``type`` (:class:`~.types.ReleaseType` enum)
      - ``title``
      - ``aka`` (Also Known As; anything after "AKA" in the title)
      - ``country``
      - ``year``
      - ``episodes`` (:class:`~.Episodes` instance)
      - ``episode_title``
      - ``date``
      - ``edition`` (:class:`list` of "Extended", "Uncut", etc)
      - ``resolution``
      - ``service`` (Streaming service abbreviation)
      - ``source`` ("BluRay", "WEB-DL", etc)
      - ``audio_codec`` (Audio codec abbreviation)
      - ``audio_channels`` (e.g. "2.0" or "7.1")
      - ``hdr_format``
      - ``video_codec``
      - ``group``
      - ``has_commentary`` (:class:`bool` or `None` to autodetect)

    Unless documented otherwise above, all values are strings. Unknown values
    are empty strings.
    """

    def __init__(self, path, strict=False):
        if strict:
            try:
                utils.predbs.assert_not_abbreviated_filename(str(path))
            except errors.SceneAbbreviatedFilenameError as e:
                raise errors.ContentError(e)
        self._path = str(path)
        self._abspath = os.path.abspath(self._path)
        self._dict = {}

    def __contains__(self, name):
        return hasattr(self, f'_get_{name}')

    def __getitem__(self, name):
        if name not in self:
            raise KeyError(name)
        else:
            value = self._dict.get(name, None)
            if value is None:
                value = self[name] = getattr(self, f'_get_{name}')()
            return value

    def __setitem__(self, name, value):
        if not hasattr(self, f'_get_{name}'):
            raise KeyError(name)
        elif hasattr(self, f'_set_{name}'):
            self._dict[name] = getattr(self, f'_set_{name}')(value)
        else:
            self._dict[name] = value

    def __delitem__(self, name):
        if not hasattr(self, f'_get_{name}'):
            raise KeyError(name)
        elif name in self._dict:
            del self._dict[name]

    def __iter__(self):
        return iter(name[5:] for name in dir(type(self))
                    if name.startswith('_get_'))

    def __len__(self):
        return len(tuple(name[5:] for name in dir(type(self))
                         if name.startswith('_get_')))

    def __repr__(self):
        return f'{type(self).__name__}({self._path!r})'

    def copy(self):
        """Return instance copy"""
        cp = type(self)(self._path)
        cp._dict = copy.deepcopy(self._dict)
        return cp

    @property
    def path(self):
        """`path` argument as :class:`str`"""
        return self._path

    @functools.cached_property
    def _guess(self):
        path = self._abspath

        # _log.debug('Running guessit on %r with %r', path, constants.GUESSIT_OPTIONS)
        guess = dict(_guessit.default_api.guessit(path, options=constants.GUESSIT_OPTIONS))
        # _log.debug('Original guess: %r', guess)

        # We try to do our own episode parsing to preserve order and support
        # multiple seasons and episodes
        for name in fs.file_and_parent(path):
            if Episodes.has_episodes_info(name):
                guess['episodes'] = Episodes.from_string(name)
                break
        else:
            # Guessit can parse multiple seasons/episodes to some degree
            seasons = _as_list(guess.get('season'))
            episodes = _as_list(guess.get('episode'))
            string = []
            if seasons:
                string.append(f'S{seasons[0]:02d}')
            for e in episodes:
                string.append(f'E{e:02d}')
            guess['episodes'] = Episodes.from_string(''.join(string))

        # If we got an abbreviated file name (e.g. "group-titles01e02.mkv"),
        # guessit can't handle it. Try to find episode information in it.
        if guess['episodes']:
            episodes_string = str(guess['episodes'])
            if (
                'S' in episodes_string
                and 'E' not in episodes_string
                and utils.predbs.is_abbreviated_filename(path)
            ):
                filename = fs.basename(path)
                match = re.search(r'((?i:[SE]\d+)+)', filename)
                if match:
                    episodes = Episodes.from_string(match.group(1))
                    guess['episodes'].update(episodes)
                    _log.debug('Found episodes in abbreviated file name: %r: %r', filename, guess['episodes'])

        return guess

    @property
    def _guessit_options(self):
        return _guessit.default_api.advanced_config

    def _get_type(self):
        if self['date']:
            # We interpret a date as an air date, which should mean it's an
            # episode.
            return ReleaseType.episode

        elif self['episodes']:
            # guessit doesn't differentiate between episodes and season packs.
            # Does any season specify an episode?
            if any(episodes for episodes in self['episodes'].values()):
                return ReleaseType.episode
            # No single episode means it's one or more season packs.
            else:
                return ReleaseType.season
        else:
            return ReleaseType.movie

    _title_split_regex = re.compile(
        r'[ \.](?:'
        r'\d{4}|'  # Year
        r'(?i:[SE]\d+)+|'  # Sxx or SxxExx
        r'((?i:Season|Episode)[ \.]*\d+[ \.]*)+|'
        r')[ \.]'
    )

    @functools.cached_property
    def release_name_params(self):
        """
        Release name without title and year or season/episode info

        This allows us to find stuff in the release name that guessit doesn't
        support without accidentally finding it in the title.
        """
        path_no_ext = fs.strip_extension(self._abspath, only=constants.VIDEO_FILE_EXTENSIONS)

        # Look for year/season/episode info in file and parent directory name
        for name in fs.file_and_parent(path_no_ext):
            match = self._title_split_regex.search(name)
            if match:
                return self._title_split_regex.split(name, maxsplit=1)[-1]

        # Default to the file/parent that contains either " " or "." (without
        # the "." from the extension)
        for name in fs.file_and_parent(path_no_ext):
            if ' ' in name or '.' in name:
                return name

        # Default to file name
        return fs.basename(path_no_ext)

    _title_aka_regex = re.compile(rf'{DELIM}AKA{DELIM}')

    @functools.cached_property
    def _title_parts(self):
        # guessit splits AKA at " - ", so we re-join it
        title_parts = [_as_string(self._guess.get('title', ''))]
        if self._guess.get('alternative_title'):
            title_parts.extend(_as_list(self._guess.get('alternative_title')))
        title = ' - '.join(title_parts)
        title_parts = self._title_aka_regex.split(title, maxsplit=1)

        # guessit recognizes mixed seasons and episodes (e.g. "S02E10S03E05") as
        # part of the title.
        def remove_episodes(string):
            return Episodes.regex.sub('', string)

        if len(title_parts) > 1:
            return {'title': remove_episodes(title_parts[0]),
                    'aka': remove_episodes(title_parts[1])}
        else:
            return {'title': remove_episodes(title_parts[0]),
                    'aka': ''}

    def _get_title(self):
        return self._title_parts['title']

    def _get_aka(self):
        return self._title_parts['aka']

    def _get_year(self):
        return _as_string(self._guess.get('year') or '')

    _country_translation = {
        'UK': re.compile(r'^GB$'),
    }

    def _get_country(self):
        country = utils.country.tld(
            _as_string(self._guess.get('country') or '')
        ).upper()
        for country_, regex in self._country_translation.items():
            if regex.search(country):
                country = country_
        return country

    def _get_episodes(self):
        if 'episodes' not in self._guess:
            self._guess['episodes'] = Episodes()
        return self._guess['episodes']

    def _set_episodes(self, value):
        # Keep Episodes() object ID. Ensure `value` is not the object in
        # self._guess['episodes'] so we can clear() without losing items in
        # `value`.
        value = dict(value)
        episodes = self._get_episodes()
        episodes.clear()
        episodes.update(value)

    def _get_episode_title(self):
        return _as_string(self._guess.get('episode_title', ''))

    _edition_translation = {
        "Collector's Edition": re.compile(r'Collector'),
        'Criterion Collection': re.compile(r'Criterion'),
        'Deluxe Edition': re.compile(r'Deluxe'),
        'Extended Cut': re.compile(r'Extended'),
        'Special Edition': re.compile(r'Special'),
        'Theatrical Cut': re.compile(r'Theatrical'),
        'Ultimate Cut': re.compile(r'Ultimate'),
    }
    _proper_repack_regex = re.compile(rf'(?:{DELIM}|^)((?i:proper|repack\d*))(?:{DELIM}|$)')
    _remastered_regex = re.compile(rf'(?:{DELIM}|^)((?i:4k{DELIM}+|)(?i:remaster(?:ed|)|restored))(?:{DELIM}|$)')

    def _get_date(self):
        return _as_string(self._guess.get('date', ''))

    def _get_edition(self):
        edition = _as_list(self._guess.get('edition'))
        for edition_fixed, regex in self._edition_translation.items():
            for i in range(len(edition)):
                if regex.search(edition[i]):
                    edition[i] = edition_fixed

        # Revision (guessit doesn't distinguish between REPACK, PROPER, etc)
        match = self._proper_repack_regex.search(self.release_name_params)
        if match:
            edition.append(match.group(1).upper())

        # Various
        guessit_other = _as_list(self._guess.get('other'))
        if 'Open Matte' in guessit_other:
            edition.append('Open Matte')
        if 'Original Aspect Ratio' in guessit_other:
            edition.append('OAR')
        if 'Dual Audio' in guessit_other:
            edition.append('Dual Audio')
        if '2in1' in guessit_other:
            edition.append('2in1')

        def is_4k_source():
            # guessit only detects remastered, not if it's from 4k
            match = self._remastered_regex.search(self.release_name_params)
            if match:
                remastered_string = match.group(1)
                return '4k' in remastered_string.lower()

        if 'Remastered' in edition and is_4k_source():
            edition[edition.index('Remastered')] = '4k Remastered'
        elif 'Restored' in edition and is_4k_source():
            edition[edition.index('Restored')] = '4k Restored'

        return edition

    def _get_resolution(self):
        return _as_string(self._guess.get('screen_size', ''))

    _streaming_service_regex = re.compile(rf'{DELIM}([A-Z]+){DELIM}(?i:WEB-?(?:DL|Rip))(?:{DELIM}|$)')
    _streaming_service_translation = {
        re.compile(r'(?i:IT)'): 'iT',
        re.compile(r'(?i:4OD)'): 'ALL4',   # 4oD is old name of Channel 4's VOD service
        # Not a streaming service
        re.compile(r'OAR'): '',  # Original Aspect Ratio
    }

    def _get_service(self):
        def translate(service):
            for regex, abbrev in self._streaming_service_translation.items():
                if regex.search(service):
                    return abbrev
            return service

        service = _as_string(self._guess.get('streaming_service', ''))
        if service:
            # guessit translates abbreviations to full names (NF -> Netflix),
            # but we want abbreviations. Use the same dictionary as guessit.
            translation = self._guessit_options['streaming_service']
            for full_name, aliases in translation.items():
                if service.casefold().strip() == full_name.casefold().strip():
                    # `aliases` is either a string or a list of strings and
                    # other objects.
                    if isinstance(aliases, str):
                        return translate(aliases)
                    else:
                        # Find shortest string
                        aliases = (a for a in aliases if isinstance(a, str))
                        return translate(sorted(aliases, key=len)[0])

        # Default to manual detection
        match = self._streaming_service_regex.search(self.release_name_params)
        if match:
            return translate(match.group(1))

        return ''

    _source_translation = {
        re.compile(r'(?i:ultra hd blu-?ray)') : 'UHD BluRay',
        re.compile(r'(?i:blu-?ray)')          : 'BluRay',
        re.compile(r'(?i:dvd-?rip)')          : 'DVDRip',
        re.compile(r'(?i:tv-?rip)')           : 'TVRip',
        re.compile(r'(?i:web-?dl)')           : 'WEB-DL',
        re.compile(r'(?i:web-?rip)')          : 'WEBRip',
        re.compile(r'(?i:web)')               : 'WEB-DL',
    }
    _web_source_regex = re.compile(rf'{DELIM}(WEB-?(?:DL|Rip))(?:{DELIM}|$)', flags=re.IGNORECASE)

    def _get_source(self):
        source = self._guess.get('source', '')

        # guessit parses multiple source (e.g. ["WEB", "Blu-ray"]) if it can
        # find them. This is an issue if an episode title contains a "source",
        # e.g. "TC" -> "Telecine".
        if not isinstance(source, str):
            source = source[0]

        if source.lower() == 'web':
            # guessit doesn't distinguish between WEB-DL and WEBRip. Get the
            # source from the path.
            match = self._web_source_regex.search(self.release_name_params)
            if match:
                source = match.group(1)

        elif 'Rip' in self._guess.get('other', ()):
            source = f'{source}Rip'

        elif 'DVD9' in self.release_name_params:
            source = 'DVD9'
        elif 'DVD5' in self.release_name_params:
            source = 'DVD5'

        if source:
            # Fix spelling
            for regex,source_fixed in self._source_translation.items():
                if regex.search(source):
                    source = source_fixed
                    break

        # Detect Remux and Hybrid
        other = self._guess.get('other', ())
        if 'Remux' in other and 'WEB' not in source and 'Rip' not in source:
            source += ' Remux'
        if 'Hybrid' in other:
            source = 'Hybrid ' + source

        return source

    _audio_codec_translation = {
        re.compile(r'^AC-?3')             : 'DD',
        re.compile(r'Dolby Digital$')     : 'DD',
        re.compile(r'^E-?AC-?3')          : 'DDP',
        re.compile(r'Dolby Digital Plus') : 'DDP',
        re.compile(r'TrueHD')             : 'TrueHD',
        re.compile(r'Dolby Atmos')        : 'Atmos',
        re.compile(r'Master Audio')       : 'MA',
        re.compile(r'High Resolution')    : 'HR',
        re.compile(r'Extended Surround')  : 'ES',
        re.compile(r'High Efficiency')    : 'HE',
        re.compile(r'Low Complexity')     : 'LC',
        re.compile(r'High Quality')       : 'HQ',
    }

    def _get_audio_codec(self):
        audio_codec = _as_string(self._guess.get('audio_codec'))
        if not audio_codec:
            return ''
        else:
            if isinstance(audio_codec, str):
                infos = [audio_codec]
            else:
                infos = audio_codec

            parts = []
            for info in infos:
                for regex,abbrev in self._audio_codec_translation.items():
                    if regex.search(info):
                        parts.append(abbrev)
                        continue

            # Join translations, default to what guessit detected
            audio_codec = ' '.join(parts or infos)

            # Final adjustments
            audio_codec = re.sub(r'DDP? TrueHD', 'TrueHD', audio_codec)
            audio_codec = re.sub(r'DDP? Atmos', 'Atmos', audio_codec)

            return audio_codec

    _audio_channels_regex = re.compile(rf'{DELIM}(\d\.\d){DELIM}')

    def _get_audio_channels(self):
        audio_channels = _as_string(self._guess.get('audio_channels', ''))
        if not audio_channels:
            match = self._audio_channels_regex.search(self.release_name_params)
            if match:
                return match.group(1)
        return audio_channels

    _hdr_translation = {
        re.compile(rf'(?:{DELIM}|^)(?i:DV|DoVi|Dolby{DELIM}Vision)(?:{DELIM}|$)'): 'DV',
        re.compile(rf'(?:{DELIM}|^)(?i:HDR10\+)(?:{DELIM}|$)'): 'HDR10+',
        re.compile(rf'(?:{DELIM}|^)(?i:HDR10)(?!:\+)(?:{DELIM}|$)'): 'HDR10',
        re.compile(rf'(?:{DELIM}|^)(?i:HDR)(?!:10\+)(?:{DELIM}|$)'): 'HDR',
    }

    def _get_hdr_format(self):
        for regex, hdr_format in self._hdr_translation.items():
            if regex.search(self.release_name_params):
                return hdr_format
        return ''

    _x264_regex = re.compile(rf'(?:{DELIM}|^)(?i:x264)(?:{DELIM}|$)')
    _x265_regex = re.compile(rf'(?:{DELIM}|^)(?i:x265)(?:{DELIM}|$)')

    def _get_video_codec(self):
        video_codec = _as_string(self._guess.get('video_codec', ''))
        if video_codec == 'H.264':
            if self._x264_regex.search(self.release_name_params):
                return 'x264'

        elif video_codec == 'H.265':
            if self._x265_regex.search(self.release_name_params):
                return 'x265'

        return video_codec

    def _get_group(self):
        return _as_string(self._guess.get('release_group', ''))

    _has_commentary_regex = re.compile(rf'{DELIM}(?i:plus{DELIM}+comm|commentary){DELIM}')

    def _get_has_commentary(self):
        if self._guess.get('has_commentary', None) is None:
            self._guess['has_commentary'] = \
                bool(self._has_commentary_regex.search(self.release_name_params))
        return self._guess['has_commentary']

    def _set_has_commentary(self, value):
        if value is None:
            self._guess['has_commentary'] = None
        else:
            self._guess['has_commentary'] = bool(value)


class Episodes(dict):
    """
    :class:`dict` subclass that maps season numbers to lists of episode numbers

    All keys and values are :class:`str` objects. All episodes from a season are
    indicated by an empty sequence. For episodes from any sason, the key is an
    empty string.

    This class accepts the same arguments as :class:`dict`.

    To provide seasons as keyword arguments, you need to prefix "S" to each
    keyword. This is because numbers can't be keyword arguments, but it also
    looks nicer.

    >>> e = Episodes({"1": ["1", "2", "3"], "2": []})
    >>> e.update(S01=[3, "4"], S3=range(2, 4), s05=[], S=[10, "E11", 12])
    >>> e
    >>> Episodes({'1': ['1', '2', '3', '4'], '2': [], '3': ['2', '3'], '5': [], '': ['10', '11', '12']})
    """

    regex = re.compile(rf'(?:{DELIM}|^)((?i:[SE]\d+)+)(?:{DELIM}|$)')
    """Regular expression that matches "S01E02"-like episode information"""

    @classmethod
    def has_episodes_info(cls, string):
        """Whether `string` contains "S01E02"-like episode information"""
        return bool(cls.regex.search(string))

    _is_episodes_info_regex = re.compile(r'^(?i:[SE]\d+)+$')

    @classmethod
    def is_episodes_info(cls, string):
        """Whether `string` is "S01E02"-like episode information and nothing else"""
        return bool(cls._is_episodes_info_regex.search(string))

    @classmethod
    def from_string(cls, value):
        """
        Create instance from release name or string that contains "Sxx" and "Exx"

        Examples:

            >>> Episodes.from_string('foo.E01 bar')
            {'': ('1',)}
            >>> Episodes.from_string('foo E01E2.bar')
            {'': ('1', '2')}
            >>> Episodes.from_string('foo.bar.E01E2S03')
            {'': ('1', '2'), '3': ()}
            >>> Episodes.from_string('E01E2S03E04E05.baz')
            {'': ('1', '2'), '3': ('4', '5')}
            >>> Episodes.from_string('S09E08S03E06S9E1')
            {'9': ('1', '8',), '3': ('6',)}
            >>> Episodes.from_string('E01S03E06.bar.E02')
            {'': ('1', '2',), '3': ('6',)}
        """
        def extract_number(string):
            n = ''.join(c for c in string if c in '0123456789').lstrip('0')
            return '0' if n == '' else n

        def split_episodes(string):
            return {extract_number(e) for e in string.split('E') if e.strip()}

        # Extract episode information to get "S..E.." string without any other
        # characters. The original value might have delimiters,
        # e.g. "...S09E14-E15...", and we want "S09E14E15".
        episodes_string = ''.join(re.findall(
            (
                r'(?i:[^SE\d]*|^)'
                r'((?i:[SE]\d+)+)'
                r'(?i:[^SE\d]+|$)'
            ),
            str(value).upper(),
        ))

        seasons = collections.defaultdict(lambda: set())
        for part in (k for k in episodes_string.split('S') if k):
            season = re.sub(r'E.*$', '', part)
            season = str(int(season)) if season else ''
            episodes = re.sub(r'^\d+', '', part)
            episodes = split_episodes(episodes) if episodes else ()
            seasons[season].update(episodes)

        args = {season: tuple(natsort.natsorted(episodes))
                for season, episodes in natsort.natsorted(seasons.items())}
        return cls(args)

    @classmethod
    def from_sequence(cls, sequence):
        """
        Combine episode information from multiple strings

        Examples:

            >>> Episodes.from_sequence(['foo.S01E01.bar', 'hello'])
            {'1': ('1',)}
            >>> Episodes.from_sequence(['foo.S01E01.bar', 'bar.S01E02.baz'])
            {'1': ('1', '2')}
        """
        episodes = Episodes()
        for string in sequence:
            eps = cls.from_string(string)
            for season in eps:
                if season in episodes:
                    episodes[season] += eps[season]
                else:
                    episodes[season] = eps[season]
        return episodes

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._set(*args, **kwargs)

    def update(self, *args, clear=False, **kwargs):
        """
        Set specific episodes from specific seasons, remove all other episodes and
        seasons

        :params bool clear: Whether to remove all seasons and episodes first
        """
        if clear:
            self.clear()
        self._set(*args, **kwargs)

    def _set(self, *args, **kwargs):
        # Validate all values before applying any changes
        validated = {}
        update = dict(*args, **kwargs)
        for season, episodes in update.items():
            season = self._normalize_season(season)
            if not isinstance(episodes, collections.abc.Iterable) or isinstance(episodes, str):
                episodes = [self._normalize_episode(episodes)]
            else:
                episodes = [self._normalize_episode(e) for e in episodes]

            if season in validated:
                validated[season].extend(episodes)
            else:
                validated[season] = episodes

        # Set validated values
        for season, episodes in validated.items():
            if season in self:
                self[season].extend(episodes)
            else:
                self[season] = episodes

            # Remove duplicates
            self[season][:] = set(self[season])

            # Sort naturally
            self[season].sort(key=natsort.natsort_key)

    def _normalize_season(self, value):
        return self._normalize(value, name='season', prefix='S', empty_string_ok=True)

    def _normalize_episode(self, value):
        return self._normalize(value, name='episode', prefix='E', empty_string_ok=False)

    def _normalize(self, value, name, prefix=None, empty_string_ok=False):
        if isinstance(value, int):
            if value >= 0:
                return str(value)

        elif isinstance(value, str):
            if value == '' and empty_string_ok:
                return str(value)

            if value.isdigit():
                return str(int(value))

            if prefix and len(value) >= len(prefix):
                prefix_ = value[:len(prefix)].casefold()
                if prefix_ == prefix.casefold():
                    actual_value = value[len(prefix):]
                    return self._normalize(
                        actual_value,
                        name=name,
                        prefix=None,
                        empty_string_ok=empty_string_ok,
                    )

        raise TypeError(f'Invalid {name}: {value!r}')

    def remove_specific_episodes(self):
        """Remove episodes from each season, leaving only complete seasons"""
        for season in tuple(self):
            if season:
                self[season] = ()
            else:
                del self[season]

    def __repr__(self):
        return f'{type(self).__name__}({dict(self)!r})'

    def __str__(self):
        parts = []
        for season, episodes in sorted(self.items()):
            if season:
                parts.append(f'S{season:0>2}')
            for episode in episodes:
                parts.append(f'E{episode:0>2}')
        return ''.join(parts)


def _as_list(value):
    if not value:
        return []
    elif isinstance(value, str):
        return [value]
    elif isinstance(value, list):
        return list(value)
    else:
        return [value]

def _as_string(value):
    if not value:
        return ''
    elif isinstance(value, str):
        return value
    elif isinstance(value, list):
        return ' '.join(value)
    else:
        return str(value)



"""
Open a URL in a web browser
"""

from . import LazyModule

import logging  # isort:skip
_log = logging.getLogger(__name__)

webbrowser = LazyModule(module='webbrowser', namespace=globals())


def open(url):
    """Attempt to open URL in default web browser"""
    webbrowser.open_new_tab(str(url))



import asyncio
import base64
import io
import itertools
import os
import pathlib
import re
import ssl
import time
from unittest.mock import AsyncMock, Mock, call

import httpx
import pytest
from pytest_httpserver.httpserver import Response
from werkzeug.datastructures import Headers, ImmutableMultiDict

from upsies import __project_name__, __version__, errors
from upsies.utils import http, semantic_hash


class RequestHandler:
    def __init__(self):
        self.requests_seen = []

    def __call__(self, request):
        # `request` is a Request object from werkzeug
        # https://werkzeug.palletsprojects.com/en/1.0.x/wrappers/#werkzeug.wrappers.BaseRequest
        try:
            return self.handle(request)
        except Exception as e:
            # pytest-httpserver doesn't show the traceback if we call
            # raise_for_status() on the response.
            import traceback
            traceback.print_exception(type(e), e, e.__traceback__)
            raise

    def handle(self, request):
        raise NotImplementedError()


@pytest.mark.parametrize(
    argnames=('cache_directory, default_cache_directory, exp_cache_directory'),
    argvalues=(
        ('my/custom/path', 'the/default/path', 'my/custom/path'),
        ('', 'the/default/path', 'the/default/path'),
        (None, 'the/default/path', 'the/default/path'),
    ),
)
@pytest.mark.asyncio
async def test_get_cache_directory(cache_directory, default_cache_directory, exp_cache_directory, mocker):
    mocker.patch('upsies.constants.DEFAULT_CACHE_DIRECTORY', default_cache_directory)
    mocker.patch('upsies.utils.http.cache_directory', cache_directory)
    assert http._get_cache_directory() == exp_cache_directory


@pytest.mark.asyncio
async def test_get(mocker):
    request_mock = mocker.patch('upsies.utils.http._request', new_callable=AsyncMock)
    kwargs = {
        'url': Mock(),
        'headers': Mock(),
        'params': Mock(),
        'auth': Mock(),
        'cache': Mock(),
        'max_cache_age': Mock(),
        'user_agent': Mock(),
        'follow_redirects': Mock(),
        'verify': Mock(),
        'timeout': Mock(),
        'cookies': Mock(),
        'debug_file': Mock(),
    }
    result = await http.get(**kwargs)
    assert request_mock.call_args_list == [
        call(**{**{'method': 'GET'}, **kwargs})
    ]
    assert result is request_mock.return_value


@pytest.mark.asyncio
async def test_post(mocker):
    request_mock = mocker.patch('upsies.utils.http._request', new_callable=AsyncMock)
    kwargs = {
        'url': Mock(),
        'headers': Mock(),
        'data': Mock(),
        'files': Mock(),
        'auth': Mock(),
        'cache': Mock(),
        'max_cache_age': Mock(),
        'user_agent': Mock(),
        'follow_redirects': Mock(),
        'verify': Mock(),
        'timeout': Mock(),
        'cookies': Mock(),
        'debug_file': Mock(),
    }
    result = await http.post(**kwargs)
    assert request_mock.call_args_list == [
        call(**{**{'method': 'POST'}, **kwargs})
    ]
    assert result is request_mock.return_value


def test_Response_is_string():
    r = http.Response('foo', b'foo')
    assert isinstance(r, str)
    assert r == 'foo'

def test_Response_bytes():
    r = http.Response('föö', bytes('föö', 'utf-8'))
    assert r.bytes == bytes('föö', 'utf-8')
    assert str(r.bytes, 'utf-8') == 'föö'

def test_Response_headers():
    r = http.Response('foo', b'foo')
    assert r.headers == {}
    r = http.Response('foo', b'foo', headers={'a': '1', 'b': '2'})
    assert r.headers == {'a': '1', 'b': '2'}

def test_Response_status_code():
    r = http.Response('foo', b'foo')
    assert r.status_code is None
    r = http.Response('foo', b'foo', status_code=304)
    assert r.status_code == 304

def test_Response_json():
    r = http.Response('{"this":"that"}', b'{"this":"that"}')
    assert r.json() == {'this': 'that'}
    r = http.Response('{"this":"that"', b'{"this":"that"')
    assert r == '{"this":"that"'
    with pytest.raises(errors.RequestError, match=r'^Malformed JSON: {"this":"that": '):
        r.json()

@pytest.mark.parametrize(
    argnames='kwargs, exp_kwargs',
    argvalues=(
        ({'text': 'föö', 'bytes': bytes('föö', 'utf-8')}, {'text': 'föö', 'bytes': bytes('föö', 'utf-8')}),
        ({'text': 'foo', 'bytes': b'foo', 'headers': {'bar': 'baz'}}, {'text': 'foo', 'bytes': b'foo', 'headers': {'bar': 'baz'}}),
        ({'text': 'foo', 'bytes': b'foo', 'headers': {}}, {'text': 'foo', 'bytes': b'foo'}),
        ({'text': 'foo', 'bytes': b'foo', 'headers': None}, {'text': 'foo', 'bytes': b'foo'}),
        ({'text': 'foo', 'bytes': b'foo', 'status_code': 123}, {'text': 'foo', 'bytes': b'foo', 'status_code': 123}),
        ({'text': 'foo', 'bytes': b'foo', 'status_code': 0}, {'text': 'foo', 'bytes': b'foo', 'status_code': 0}),
        ({'text': 'foo', 'bytes': b'foo', 'status_code': None}, {'text': 'foo', 'bytes': b'foo'}),
    ),
    ids=lambda v: str(v),
)
def test_Response_repr(kwargs, exp_kwargs):
    r = http.Response(**kwargs)
    exp_kwargs_string = ', '.join(f'{k}={v!r}' for k, v in exp_kwargs.items())
    assert repr(r) == f'Response({exp_kwargs_string})'


@pytest.mark.asyncio
async def test_request_with_invalid_url(mock_cache):
    # The error message keeps changing, so we just make sure RequestError is raised
    url = r'http:///baz'
    with pytest.raises(errors.RequestError, match=(rf"^{re.escape(url)}: ")) as excinfo:
        await http._request('GET', url)
    assert excinfo.value.status_code is None


@pytest.mark.asyncio
async def test_request_with_invalid_method(mock_cache):
    with pytest.raises(ValueError, match=r'^Invalid method: hello$'):
        await http._request('hello', 'asdf')


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_with_caching_disabled(method, mock_cache, httpserver):
    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_data(
        'have this',
    )
    for i in range(3):
        result = await http._request(
            method=method,
            url=httpserver.url_for('/foo'),
            cache=False,
        )
        assert result == 'have this'
        assert isinstance(result, http.Response)
        assert mock_cache.mock_calls == []


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_gets_cached_result(method, mock_cache):
    mock_cache.from_cache.return_value = 'cached result'
    url = 'http://localhost:12345/foo'
    for i in range(1, 4):
        result = await http._request(method=method, url=url, cache=True)
        assert result == 'cached result'
        assert result is mock_cache.from_cache.return_value
        assert mock_cache.mock_calls == [
            call.cache_file(method, url, {}),
            call.from_cache(mock_cache.cache_file.return_value, max_age=float('inf')),
        ] * i


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_caches_result(method, mock_cache, httpserver):
    mock_cache.from_cache.return_value = None
    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_data(
        'have this',
    )
    result = await http._request(
        method=method,
        url=httpserver.url_for('/foo'),
        cache=True,
    )
    assert result == 'have this'
    assert isinstance(result, http.Response)
    assert mock_cache.mock_calls == [
        call.cache_file(method, httpserver.url_for('/foo'), {}),
        call.from_cache(mock_cache.cache_file.return_value, max_age=float('inf')),
        call.cache_file(method, httpserver.url_for('/foo'), {}),
        call.to_cache(mock_cache.cache_file.return_value, b'have this'),
    ]


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_sends_default_headers(method, mock_cache, httpserver):
    class Handler(RequestHandler):
        def handle(self, request):
            for k, v in http._default_headers.items():
                assert request.headers[k] == v
            return Response('have this')

    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_handler(
        Handler(),
    )
    result = await http._request(
        method=method,
        url=httpserver.url_for('/foo'),
        user_agent=True,
    )
    assert result == 'have this'
    assert isinstance(result, http.Response)


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_sends_custom_headers(method, mock_cache, httpserver, mocker):
    mocker.patch.dict(http._default_headers, {'a': '1', 'b': '2'})
    custom_headers = {'foo': 'bar', 'b': '20'}
    combined_headers = {'a': '1', 'b': '20', 'foo': 'bar'}

    class Handler(RequestHandler):
        def handle(self, request):
            for k, v in combined_headers.items():
                assert request.headers[k] == v
            return Response('have this')

    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_handler(
        Handler(),
    )
    result = await http._request(
        method=method,
        url=httpserver.url_for('/foo'),
        headers=custom_headers,
    )
    assert result == 'have this'
    assert isinstance(result, http.Response)


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_sends_params(method, mock_cache, httpserver):
    query = {'foo': '1', 'bar': 'baz'}
    httpserver.expect_request(
        uri='/foo',
        method=method,
        query_string=query,
    ).respond_with_data(
        'have this',
    )
    result = await http._request(
        method=method,
        url=httpserver.url_for('/foo'),
        params=query,
    )
    assert result == 'have this'
    assert isinstance(result, http.Response)


@pytest.mark.parametrize(
    argnames='data, files, exp_data, exp_form',
    argvalues=(
        (b'raw bytes', {}, b'raw bytes', ImmutableMultiDict([])),
        ('string', {}, b'string', ImmutableMultiDict([])),
        (b'raw bytes', {'file_input': {'file': io.BytesIO(b'foo')},}, b'raw bytes', ImmutableMultiDict([])),
        ('string', {'file_input': {'file': io.BytesIO(b'foo')}}, b'string', ImmutableMultiDict([])),
        (
            {'foo': 1, 'bar': None, 'baz': [1, 2, 3],},
            {},
            b'foo=1&baz=1&baz=2&baz=3',
            ImmutableMultiDict([
                ('foo', '1'),
                ('baz', '1'), ('baz', '2'), ('baz', '3'),
            ]),
        ),
        (
            {'foo': 1, 'bar': None, 'baz': [1, 2, 3],},
            {
                'file_input': {
                    'file': io.BytesIO(b'foo'),
                },
            },
            # Encoded multipart/form-data is impossible to predict
            None,
            ImmutableMultiDict([
                ('foo', '1'),
                ('baz', '1'), ('baz', '2'), ('baz', '3'),
                ('file_input', 'foo'),
            ]),
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_sends_data(method, data, files, exp_data, exp_form, mock_cache, httpserver):
    class Handler(RequestHandler):
        def handle(self, request):
            if exp_data is not None:
                assert request.data == exp_data
            if exp_form is not None:
                assert request.form == exp_form
            return Response('have this')

    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_handler(
        Handler()
    )
    result = await http._request(
        method=method,
        url=httpserver.url_for('/foo'),
        data=data,
        files=files,
    )
    assert result == 'have this'
    assert isinstance(result, http.Response)


@pytest.mark.asyncio
async def test_request_sends_files(mock_cache, httpserver, mocker):
    files = {
        'foo': 'path/to/foo.jpg',
        'bar': ('path/to/bar', 'image/png'),
    }
    mocker.patch('upsies.utils.http._open_files', return_value={
        'foo': ('foo.jpg', io.BytesIO(b'foo image')),
        'bar': ('bar', io.BytesIO(b'bar image'), 'image/png')
    })

    class Handler(RequestHandler):
        def handle(self, request):
            assert request.content_type.startswith('multipart/form-data')
            data = request.data.decode('utf-8')
            assert re.search(
                (
                    r'^'
                    r'--[0-9a-f]{32}\r\n'
                    r'Content-Disposition: form-data; name="foo"; filename="foo.jpg"\r\n'
                    r'Content-Type: image/jpeg\r\n'
                    r'\r\n'
                    r'foo image\r\n'
                    r'--[0-9a-f]{32}\r\n'
                    r'Content-Disposition: form-data; name="bar"; filename="bar"\r\n'
                    r'Content-Type: image/png\r\n'
                    r'\r\n'
                    r'bar image\r\n'
                    r'--[0-9a-f]{32}--\r\n'
                    r'$'
                ),
                data,
            )
            return Response('have this')

    httpserver.expect_request(
        uri='/foo',
        method='POST',
    ).respond_with_handler(
        Handler(),
    )
    result = await http._request(
        method='POST',
        url=httpserver.url_for('/foo'),
        files=files,
    )
    assert result == 'have this'
    assert isinstance(result, http.Response)


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_sends_auth(method, mock_cache, httpserver):
    auth = ('AzureDiamond', 'hunter2')
    auth_str = ':'.join(auth)
    auth_encoded = base64.b64encode(auth_str.encode('utf-8')).decode()

    class Handler(RequestHandler):
        def handle(self, request):
            assert request.headers['Authorization'] == f'Basic {auth_encoded}'
            return Response('have this')

    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_handler(
        Handler(),
    )
    result = await http._request(
        method=method,
        url=httpserver.url_for('/foo'),
        auth=auth,
    )
    assert result == 'have this'
    assert isinstance(result, http.Response)


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_does_not_send_auth(method, mock_cache, httpserver):
    class Handler(RequestHandler):
        def handle(self, request):
            assert request.headers.get('Authorization') is None
            return Response('have this')

    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_handler(
        Handler(),
    )
    result = await http._request(
        method=method,
        url=httpserver.url_for('/foo'),
        auth=None,
    )
    assert result == 'have this'
    assert isinstance(result, http.Response)


@pytest.mark.parametrize(
    argnames='user_agent, exp_user_agent',
    argvalues=(
        (True, f'{__project_name__}/{__version__}'),
        (False, None),
        ('Mozilla/5.0 (compatible; MSIE 8.0; Windows NT 6.3; Win64; x64)',
         'Mozilla/5.0 (compatible; MSIE 8.0; Windows NT 6.3; Win64; x64)'),
    ),
)
@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_sends_user_agent(method, user_agent, exp_user_agent, mock_cache, httpserver):
    class Handler(RequestHandler):
        def handle(self, request):
            assert request.headers.get('User-Agent') == exp_user_agent
            return Response('have this')

    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_handler(
        Handler(),
    )
    result = await http._request(
        method=method,
        url=httpserver.url_for('/foo'),
        user_agent=user_agent,
    )
    assert result == 'have this'
    assert isinstance(result, http.Response)


@pytest.mark.parametrize(
    argnames='timeout, response_delay, exp_exception',
    argvalues=(
        (1, 0.5, None),
        (1, 1.5, errors.RequestError('timeout')),
    ),
)
@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_with_timeout_argument(method, timeout, response_delay, exp_exception, mock_cache, httpserver):
    class Handler(RequestHandler):
        def handle(self, request):
            import time
            time.sleep(response_delay)
            return Response('have this')

    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_handler(
        Handler(),
    )
    request_url = httpserver.url_for('/foo')
    if exp_exception:
        with pytest.raises(errors.RequestError, match=rf'^{request_url}: Timeout$'):
            await http._request(
                method=method,
                url=request_url,
                timeout=timeout,
            )
    else:
        result = await http._request(
            method=method,
            url=request_url,
            timeout=timeout,
        )
        assert result == 'have this'
        assert isinstance(result, http.Response)


@pytest.mark.parametrize('follow_redirects', (True, False))
@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.parametrize('status_code', (301, 302, 303, 307, 308))
@pytest.mark.asyncio
async def test_request_with_follow_redirects(status_code, method, follow_redirects, mock_cache, httpserver):
    class Handler(RequestHandler):
        def handle(self, request):
            if request.path == '/foo':
                return Response(
                    response='not redirected',
                    status=status_code,
                    headers={'Location': httpserver.url_for('/bar')},
                )
            else:
                return Response('redirected')

    handler = Handler()
    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_handler(
        handler,
    )
    httpserver.expect_request(
        uri='/bar',
        method='GET' if status_code in (301, 302, 303) else method,
    ).respond_with_handler(
        handler,
    )

    result = await http._request(
        method=method,
        url=httpserver.url_for('/foo'),
        follow_redirects=follow_redirects,
    )
    if follow_redirects:
        assert result == 'redirected'
    else:
        assert result == 'not redirected'
    assert isinstance(result, http.Response)


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_preserves_session_cookies_between_requests(method, mock_cache, httpserver):
    class Handler(RequestHandler):
        expected_cookies = ImmutableMultiDict({})

        def handle(self, request):
            print('SERVER: Got cookies:', request.cookies)
            assert request.cookies == self.expected_cookies

            bar = int(request.cookies.get('bar', 0))
            self.expected_cookies = ImmutableMultiDict({'bar': str(bar + 1), 'baz': 'asdf'})

            if not request.cookies:
                # First request
                headers = Headers((
                    ('Set-Cookie', f'bar={bar + 1}'),
                    ('Set-Cookie', 'baz=asdf'),
                ))
            else:
                headers = Headers((
                    ('Set-Cookie', f'bar={bar + 1}'),
                ))
            print('SERVER: Sending cookies:', headers)

            return Response(response=f'bar is currently {bar}', headers=headers)

    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_handler(
        Handler(),
    )

    try:
        for i in range(3):
            response = await http._request(
                method=method,
                url=httpserver.url_for('/foo'),
                cache=False,
            )
            assert response == f'bar is currently {i}'
    finally:
        http.clear_session_cookies()


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_uses_separate_cookie_jar_per_domain(method, mock_cache, httpserver):
    class Handler(RequestHandler):
        def handle(self, request):
            if getattr(self, 'set_cookie'):
                domain = request.url.split('/')[2]
                headers = {'Set-Cookie': f'cookie_domain={domain}'}
            else:
                headers = {}
            return Response(
                response='current cookies: ' + ', '.join(f'{k}={v}' for k, v in request.cookies.items()),
                headers=headers,
            )

    handler = Handler()
    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_handler(
        handler,
    )

    hosts = ('foo.localhost', 'bar.localhost', 'baz.localhost')

    try:
        # Fill cookie jar with some subdomain-specific cookies
        handler.set_cookie = True
        for host in hosts:
            httpserver.host = host
            response = await http._request(
                method=method,
                url=httpserver.url_for('/foo'),
                cache=False,
            )
            assert response == 'current cookies: '

        # Check if we get the same cookies back for each subdomain
        handler.set_cookie = False
        for host in hosts:
            httpserver.host = host
            response = await http._request(
                method=method,
                url=httpserver.url_for('/foo'),
                cache=False,
            )
            assert response == f'current cookies: cookie_domain={host}:{httpserver.port}'

    finally:
        http.clear_session_cookies()


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_with_custom_cookies(method, mock_cache, httpserver):
    custom_cookies = {'foo': 'bar'}
    server_cookies = {'server': 'cookie'}

    class Handler(RequestHandler):
        def handle(self, request):
            if getattr(self, 'cookies_already_set', False):
                headers = {}
                exp_request_cookies = ImmutableMultiDict({**custom_cookies, **server_cookies})
            else:
                headers = {'Set-Cookie': ', '.join(f'{k}={v}' for k, v in server_cookies.items())}
                self.cookies_already_set = True
                exp_request_cookies = ImmutableMultiDict(custom_cookies)

            assert request.cookies == exp_request_cookies

            return Response(
                response='got cookies: ' + ', '.join(f'{k}={v}' for k, v in request.cookies.items()),
                headers=headers,
            )

    handler = Handler()
    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_handler(
        handler,
    )

    try:
        response = await http._request(
            method=method,
            url=httpserver.url_for('/foo'),
            cache=False,
            cookies=custom_cookies,
        )
        assert response == 'got cookies: ' + ', '.join(f'{k}={v}' for k, v in custom_cookies.items())

        response = await http._request(
            method=method,
            url=httpserver.url_for('/foo'),
            cache=False,
            cookies=custom_cookies,
        )
        assert response == 'got cookies: ' + ', '.join(f'{k}={v}' for k, v in {**custom_cookies, **server_cookies}.items())

    finally:
        http.clear_session_cookies()


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_with_cookies_file(method, mock_cache, httpserver, custom_host, tmp_path):
    cookies_filepath = str(tmp_path / 'path' / 'to' / 'my.cookies')

    class Handler(RequestHandler):
        def handle(self, request):
            assert request.cookies == ImmutableMultiDict(self.expected_cookies)

            domain = request.url.split('/')[2].split(':')[0]
            if getattr(self, 'set_cookie'):
                headers = {'Set-Cookie': f'your_domain={domain}; max-age=500000'}
            else:
                headers = {}
            domain = request.url.split('/')[2]
            return Response(
                response=f'got cookies on {domain}: ' + ', '.join(f'{k}={v}' for k, v in request.cookies.items()),
                headers=headers,
            )

    handler = Handler()
    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_handler(
        handler,
    )

    # We are only interested in cookies sourced from file
    http.clear_session_cookies()

    # Set cookies
    for domain in ('foo.localhost', 'bar.localhost'):
        with custom_host(httpserver, domain):
            handler.set_cookie = True
            handler.expected_cookies = {}
            response = await http._request(
                method=method,
                url=httpserver.url_for('/foo'),
                cache=False,
                cookies=cookies_filepath + f'.{domain}',
            )
            assert response == f'got cookies on {domain}:{httpserver.port}: '

    # We are only interested in cookies sourced from file
    http.clear_session_cookies()

    # Check if we only send the cookies for each domain
    for domain in ('foo.localhost', 'bar.localhost'):
        with custom_host(httpserver, domain):
            handler.set_cookie = False
            handler.expected_cookies = {'your_domain': domain}
            response = await http._request(
                method=method,
                url=httpserver.url_for('/foo'),
                cache=False,
                cookies=cookies_filepath + f'.{domain}',
            )
            assert response == f'got cookies on {domain}:{httpserver.port}: your_domain={domain}'


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_with_unsavable_cookies_file(method, mock_cache, httpserver, tmp_path, mocker):
    class Handler(RequestHandler):
        def handle(self, request):
            headers = {'Set-Cookie': 'your_cookie=foo; max-age=500000'}
            return Response(
                response='setting cookie',
                headers=headers,
            )

    handler = Handler()
    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_handler(
        handler,
    )

    httpserver.host = 'foo.localhost'
    cookies_filepath = tmp_path / 'unwritable' / 'my.cookies'
    cookies_filepath.parent.mkdir()

    def mkdir(path):
        pathlib.Path(path).chmod(0o000)

    mocker.patch('upsies.utils.fs.mkdir', Mock(side_effect=mkdir))

    try:
        with pytest.raises(errors.RequestError, match=rf'^Failed to write {cookies_filepath}: Permission denied$'):
            await http._request(
                method=method,
                url=httpserver.url_for('/foo'),
                cache=False,
                cookies=cookies_filepath,
            )
    finally:
        cookies_filepath.parent.chmod(0o600)


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_with_unloadable_cookies_file(method, mock_cache, tmp_path):
    cookies_filepath = tmp_path / 'my.cookies'
    cookies_filepath.write_text('mock cookies')
    cookies_filepath.chmod(0o000)
    try:
        with pytest.raises(errors.RequestError, match=rf'^Failed to read {cookies_filepath}: Permission denied$'):
            await http._request(
                method=method,
                url='http://localhost:123/foo',
                cache=False,
                cookies=cookies_filepath,
            )
    finally:
        cookies_filepath.chmod(0o600)


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_with_unsupported_cookies_type(method, mock_cache):
    with pytest.raises(RuntimeError, match=r'^Unsupported cookies type: \(1, 2, 3\)$'):
        await http._request(
            method=method,
            url='http://localhost:123/foo',
            cache=False,
            cookies=(1, 2, 3),
        )


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_catches_HTTP_error_status(method, mock_cache, httpserver):
    url = httpserver.url_for('/foo')
    httpserver.expect_request(
        uri='/foo',
        method=method,
    ).respond_with_data(
        '<html>  Dave is not   here\n</html> ',
        status=404,
        headers={'foo': 'bar'},
    )
    with pytest.raises(errors.RequestError, match=rf'^{url}: Dave is not here$') as excinfo:
        await http._request(method=method, url=url)
    assert excinfo.value.status_code == 404
    assert excinfo.value.headers['foo'] == 'bar'
    assert excinfo.value.url == url
    assert excinfo.value.text == '<html>  Dave is not   here\n</html> '


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_catches_NetworkError(method, mock_cache):
    url = 'http://localhost:12345/foo/bar/baz'
    with pytest.raises(errors.RequestError, match=rf'^{url}: All connection attempts failed$') as excinfo:
        await http._request(method=method, url=url)
    assert excinfo.value.status_code is None
    assert excinfo.value.headers == {}


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_catches_TimeoutException(method, mock_cache, mocker):
    exc = httpx.TimeoutException(
        message='Some error',
        request='mock request',
    )
    mocker.patch('httpx.AsyncClient.send', AsyncMock(side_effect=exc))
    url = 'http://localhost:12345/foo/bar/baz'
    with pytest.raises(errors.RequestError, match=rf'^{url}: Timeout$') as excinfo:
        await http._request(method=method, url=url)
    assert excinfo.value.status_code is None
    assert excinfo.value.headers == {}


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.asyncio
async def test_request_catches_HTTPError(method, mock_cache, mocker):
    exc = httpx.HTTPError('Some error')
    mocker.patch('httpx.AsyncClient.send', AsyncMock(side_effect=exc))
    url = 'http://localhost:12345/foo/bar/baz'
    with pytest.raises(errors.RequestError, match=rf'^{url}: Some error$') as excinfo:
        await http._request(method=method, url=url)
    assert excinfo.value.status_code is None
    assert excinfo.value.headers == {}


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.parametrize(
    argnames='exception, exception_attrs, exp_exception',
    argvalues=(
        (
            ssl.SSLError('Default message'),
            {'verify_message': 'Verification failed'},
            errors.RequestError('TLS: Verification failed'),
        ),
        (
            ssl.SSLError('Default message'),
            {'reason': 'Verification failed'},
            errors.RequestError('TLS: Verification failed'),
        ),
        (
            ssl.SSLError('Default message'),
            {'strerror': 'Verification failed'},
            errors.RequestError('TLS: Verification failed'),
        ),
        (
            ssl.SSLError('Default message'),
            {},
            errors.RequestError("TLS: ('Default message',)"),
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_request_catches_SSLError(exception, exception_attrs, exp_exception, method, mock_cache, mocker):
    # It seems to be impossible to normally specify `verify_message` and it is
    # not documented how to instantiate SSLCertVerificationError anywhere.
    for attr, value in exception_attrs.items():
        setattr(exception, attr, value)
    mocker.patch('httpx.AsyncClient.send', AsyncMock(side_effect=exception))
    url = 'https://localhost:12345/foo/bar/baz'
    with pytest.raises(type(exp_exception), match=rf'^{url}: {re.escape(str(exp_exception))}$') as excinfo:
        await http._request(method=method, url=url)
    assert excinfo.value.status_code is None
    assert excinfo.value.headers == {}


@pytest.mark.parametrize('debug_file', ('', None, 'path/to/debug.file'))
@pytest.mark.asyncio
async def test_request_with_debug_file_argument(debug_file, mocker):
    mock_request = Mock(
        headers={'User-Agent': 'asdf'},
        aread=AsyncMock(),
    )
    mocker.patch('httpx.AsyncClient.build_request', Mock(return_value=mock_request))
    mock_response = Mock()
    mocker.patch('httpx.AsyncClient.send', AsyncMock(return_value=mock_response))

    dump_req_res_mock = mocker.patch('upsies.utils.http._dump_req_res', AsyncMock())
    await http._request('GET', 'http://foo.bar', debug_file=debug_file)
    if debug_file:
        assert dump_req_res_mock.call_args_list == [
            call(mock_request, f'{debug_file}.request'),
            call(mock_response, f'{debug_file}.response'),
        ]
    else:
        assert dump_req_res_mock.call_args_list == []


@pytest.mark.asyncio
async def test_get_performs_only_one_identical_request_at_the_same_time(httpserver, mocker, tmp_path):
    mocker.patch.object(http, 'cache_directory', str(tmp_path))

    class Handler(RequestHandler):
        def handle(self, request):
            self.requests_seen.append(request.full_path)
            return Response(request.full_path)

    handler = Handler()
    for path in ('/a', '/b', '/c', '/d'):
        httpserver.expect_request(
            uri=path,
            method='GET',
        ).respond_with_handler(
            handler,
        )

    coros = tuple(itertools.chain(
        (http.get(httpserver.url_for('/a'), cache=True) for _ in range(10)),
        (http.get(httpserver.url_for('/b'), cache=True) for _ in range(5)),
        (http.get(httpserver.url_for('/c'), params={'foo': '1'}, cache=True) for _ in range(3)),
        (http.get(httpserver.url_for('/c'), params={'foo': '2'}, cache=True) for _ in range(6)),
    ))
    results = await asyncio.gather(*coros)

    assert set(results) == set(
        ['/a?'] * 10
        + ['/b?'] * 5
        + ['/c?foo=1'] * 3
        + ['/c?foo=2'] * 6
    )
    assert set(handler.requests_seen) == {
        '/a?',
        '/b?',
        '/c?foo=1',
        '/c?foo=2',
    }

@pytest.mark.asyncio
async def test_post_request_performs_only_one_identical_request_at_the_same_time(httpserver, mocker, tmp_path):
    mocker.patch.object(http, 'cache_directory', str(tmp_path))

    class Handler(RequestHandler):
        def handle(self, request):
            s = f'{request.full_path} data:{request.data.decode("utf-8")}'
            self.requests_seen.append(s)
            return Response(s)

    handler = Handler()
    for path in ('/a', '/b', '/c', '/d'):
        httpserver.expect_request(
            uri=path,
            method='POST',
        ).respond_with_handler(
            handler,
        )

    coros = tuple(itertools.chain(
        (http.post(httpserver.url_for('/a'), cache=True) for _ in range(10)),
        (http.post(httpserver.url_for('/b'), cache=True) for _ in range(5)),
        (http.post(httpserver.url_for('/c'), data={'foo': '1'}, cache=True) for _ in range(3)),
        (http.post(httpserver.url_for('/c'), data={'foo': '2'}, cache=True) for _ in range(6)),
    ))
    results = await asyncio.gather(*coros)

    assert set(results) == set(
        ["/a? data:"] * 10
        + ['/b? data:'] * 5
        + ['/c? data:foo=1'] * 3
        + ['/c? data:foo=2'] * 6
    )
    assert len(results) == 10 + 5 + 3 + 6
    assert set(handler.requests_seen) == {
        '/a? data:',
        '/b? data:',
        '/c? data:foo=1',
        '/c? data:foo=2',
    }
    assert len(handler.requests_seen) == 4


@pytest.mark.asyncio
async def test_download_forwards_args_and_kwargs_to_get(mocker, tmp_path):
    mocker.patch('upsies.utils.http.get', AsyncMock(
        return_value=Mock(bytes=b'downloaded data'),
    ))
    filepath = tmp_path / 'downloaded'
    return_value = await http.download('mock url', filepath, 'foo', bar='baz')
    assert return_value == filepath
    assert http.get.call_args_list == [call('mock url', 'foo', bar='baz', cache=False)]

@pytest.mark.parametrize(
    argnames='cache, file_exists, exp_download',
    argvalues=(
        (False, False, True),
        (False, True, True),
        (True, False, True),
        (True, True, False),
    ),
)
@pytest.mark.asyncio
async def test_download_with_cache_argument(cache, file_exists, exp_download, mocker, tmp_path):
    mocker.patch('upsies.utils.http.get', AsyncMock(
        return_value=Mock(bytes=b'downloaded data'),
    ))
    filepath = tmp_path / 'downloaded'
    if file_exists:
        filepath.write_bytes(b'downloaded data')
    return_value = await http.download('mock url', filepath, cache=cache)
    assert return_value == filepath
    if exp_download:
        assert http.get.call_args_list == [call('mock url', cache=False)]
    else:
        assert http.get.call_args_list == []

@pytest.mark.asyncio
async def test_download_writes_filepath(mocker, tmp_path):
    mocker.patch('upsies.utils.http.get', AsyncMock(
        side_effect=errors.RequestError('no response'),
    ))
    filepath = tmp_path / 'downloaded'
    with pytest.raises(errors.RequestError, match=r'^no response$'):
        await http.download('mock url', filepath)
    assert not filepath.exists()

@pytest.mark.asyncio
async def test_download_does_not_write_filepath_if_request_fails(mocker, tmp_path):
    mocker.patch('upsies.utils.http.get', AsyncMock(
        return_value=Mock(bytes=b'downloaded data'),
    ))
    filepath = tmp_path / 'downloaded'
    return_value = await http.download('mock url', filepath)
    assert return_value == filepath
    assert filepath.read_bytes() == b'downloaded data'

@pytest.mark.asyncio
async def test_download_catches_OSError_when_opening_filepath(mocker, tmp_path):
    mocker.patch('upsies.utils.http.get', AsyncMock(
        return_value=Mock(bytes=b'downloaded data'),
    ))
    filepath = tmp_path / 'downloaded'
    mocker.patch('builtins.open', side_effect=OSError('Ouch'))
    with pytest.raises(errors.RequestError, match=rf'^Unable to write {filepath}: Ouch$'):
        await http.download('mock url', filepath)


@pytest.mark.parametrize('domain', (None, 'example.org'))
def test_clear_session_cookies(domain, mocker):
    mocker.patch.object(http, '_session_cookies', {
        'foo.net': {'cookie1': 'bar', 'cookie2': 'baz'},
        'example.org': {'cookie2': 'this', 'cookie3': 'that'},
    })
    http.clear_session_cookies(domain)
    if domain:
        assert http._session_cookies == {
            'foo.net': {'cookie1': 'bar', 'cookie2': 'baz'},
            'example.org': {},
        }
    else:
        assert http._session_cookies == {
            'foo.net': {},
            'example.org': {},
        }


def test_open_files_opens_files(mocker):
    mocker.patch(
        'upsies.utils.http._get_file_object',
        side_effect=(
            io.BytesIO(b'foo image'),
            io.BytesIO(b'bar text1'),
            io.BytesIO(b'bar text2'),
            io.BytesIO(b'bar text3'),
            io.BytesIO(b'bar text4'),
        ),
    )
    opened_files = http._open_files({
        'a': 'path/to/foo.jpg',
        'b': {'file': 'path/to/bar.txt'},
        'c': {'file': 'path/to/bar.txt', 'filename': 'baz'},
        'd': {'file': 'path/to/bar.txt', 'mimetype': 'text/plain'},
        'e': {'file': 'path/to/bar.txt', 'mimetype': None},
        'f': io.BytesIO(b'in-memory data1'),
        'g': {'file': io.BytesIO(b'in-memory data2'), 'mimetype': 'weird/type'},
        'h': {'file': io.BytesIO(b'in-memory data3'), 'mimetype': 'weird/type', 'filename': 'kangaroo'},
        'i': {'file': io.BytesIO(b'in-memory data4'), 'mimetype': None, 'filename': 'kangaroo'},
        'j': {'file': io.BytesIO(b'in-memory data5'), 'filename': 'kangaroo'},
    })
    assert tuple(opened_files) == ('a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j')

    def assert_opened(fieldname, filename, data, mimetype=False):
        assert opened_files[fieldname][0] == filename
        assert opened_files[fieldname][1].read() == data
        if mimetype is False:
            assert len(opened_files[fieldname]) == 2
        else:
            assert opened_files[fieldname][2] == mimetype
            assert len(opened_files[fieldname]) == 3

    assert_opened('a', 'foo.jpg', b'foo image')
    assert_opened('b', 'bar.txt', b'bar text1')
    assert_opened('c', 'baz', b'bar text2')
    assert_opened('d', 'bar.txt', b'bar text3', 'text/plain')
    assert_opened('e', 'bar.txt', b'bar text4', None)
    assert_opened('f', None, b'in-memory data1')
    assert_opened('g', None, b'in-memory data2', 'weird/type')
    assert_opened('h', 'kangaroo', b'in-memory data3', 'weird/type')
    assert_opened('i', 'kangaroo', b'in-memory data4', None)
    assert_opened('j', 'kangaroo', b'in-memory data5')

def test_open_files_catches_invalid_file_value(mocker):
    mocker.patch('upsies.utils.http._get_file_object', return_value='mock file object')
    with pytest.raises(RuntimeError, match=r'^Invalid "file" value in fileinfo: \[1, 2, 3\]$'):
        http._open_files({
            'document': {'file': [1, 2, 3]},
        })

def test_open_files_catches_invalid_fileinfo(mocker):
    mocker.patch('upsies.utils.http._get_file_object', return_value='mock file object')
    with pytest.raises(RuntimeError, match=r'^Invalid fileinfo: \[1, 2, 3\]$'):
        http._open_files({
            'document': [1, 2, 3],
        })

def test_open_files_raises_exception_from_get_file_object(mocker):
    mocker.patch('upsies.utils.http._get_file_object',
                 side_effect=errors.RequestError('bad io'))
    with pytest.raises(errors.RequestError, match=r'^bad io$'):
        http._open_files({
            'image': 'path/to/foo.jpg',
            'document': ('path/to/bar', 'text/plain'),
        })


def test_get_file_object_catches_OSError():
    filepath = 'path/to/foo'
    with pytest.raises(errors.RequestError, match=rf'^{filepath}: No such file or directory$'):
        http._get_file_object(filepath)

def test_get_file_object_returns_fileobj(tmp_path):
    filepath = tmp_path / 'foo'
    filepath.write_bytes(b'hello')
    fileobj = http._get_file_object(filepath)
    assert fileobj.read() == b'hello'


@pytest.mark.parametrize('method', ('GET', 'POST'))
def test_cache_file_without_params(method, mocker):
    def sanitize_filename(filename):
        return filename.replace('a', '#')

    mocker.patch('upsies.utils.fs.sanitize_filename', side_effect=sanitize_filename)
    mocker.patch.object(http, '_get_cache_directory', return_value='/tmp/foo')
    url = 'http://localhost:123/foo/bar'
    exp_cache_file = f'/tmp/foo/{method.upper()}.http://loc#lhost:123/foo/b#r'
    assert http._cache_file(method, url) == exp_cache_file

@pytest.mark.parametrize('method', ('GET', 'POST'))
def test_cache_file_with_params(method, mocker):
    def sanitize_filename(filename):
        return filename.replace('a', '#')

    mocker.patch('upsies.utils.fs.sanitize_filename', side_effect=sanitize_filename)
    mocker.patch.object(http, '_get_cache_directory', return_value='/tmp/foo')
    url = 'http://localhost:123'
    params = {'foo': 'a b c', 'bar': 12}
    exp_cache_file = f'/tmp/foo/{method.upper()}.http://loc#lhost:123?foo=#+b+c&b#r=12'
    assert http._cache_file(method, url, params=params) == exp_cache_file

@pytest.mark.parametrize('method', ('GET', 'POST'))
def test_cache_file_with_very_long_params(method, mocker):
    def sanitize_filename(filename):
        return filename.replace('a', '#')

    mocker.patch('upsies.utils.fs.sanitize_filename', side_effect=sanitize_filename)
    mocker.patch.object(http, '_get_cache_directory', return_value='/tmp/foo')
    url = 'http://localhost:123'
    params = {'foo': 'a b c ' * 100, 'bar': 12}
    exp_cache_file = (
        f'/tmp/foo/{method.upper()}.http://loc#lhost:123?'
        f'[HASH:{sanitize_filename(semantic_hash(params))}]'
    )
    assert http._cache_file(method, url, params=params) == exp_cache_file


def test_from_cache_with_nonexisting_cache_file():
    assert http._from_cache('/no/such/path') is None

def test_from_cache_with_unreadable_cache_file(tmp_path):
    cache_filepath = tmp_path / 'cache_file'
    cache_filepath.write_text('mock data')
    cache_filepath.chmod(0o000)
    try:
        assert http._from_cache(cache_filepath) is None
    finally:
        cache_filepath.chmod(0o600)

@pytest.mark.parametrize(
    argnames='cached_data, max_age, cache_file_age, exp_return_value',
    argvalues=(
        ('föö', 100, 99, http.Response('föö', b'f\xc3\xb6\xc3\xb6')),
        ('föö', 100, 101, None),
    ),
)
def test_from_cache_refuses_to_read_old_cache_file(cached_data, max_age, cache_file_age, exp_return_value, tmp_path):
    cache_filepath = tmp_path / 'cache/file'
    cache_filepath.parent.mkdir(parents=True)
    cache_filepath.write_text(cached_data, encoding='UTF-8')
    mtime = time.time() - cache_file_age
    os.utime(cache_filepath, times=(mtime, mtime))
    cached_result = http._from_cache(cache_filepath, max_age=max_age)
    assert cached_result == exp_return_value


def test_to_cache_requires_bytes_object(tmp_path):
    cache_file = str(tmp_path / 'cachefile')
    with pytest.raises(TypeError, match=r"^Not a bytes object: 'foo'$"):
        http._to_cache(cache_file, 'foo')
    assert not os.path.exists(cache_file)

@pytest.mark.parametrize(
    argnames='bytes, exp_cached_bytes',
    argvalues=(
        (b'foo <script>bar</script> baz', b'foo  baz'),
        (b'foo <script with="attribute">bar</script> baz', b'foo  baz'),
        (b'a < b && b > c', b'a < b && b > c'),
        (b'\xc3\x28 <script', b'\xc3\x28 <script'),
    ),
)
def test_to_cache_removes_javascript_if_possible(bytes, exp_cached_bytes, mocker, tmp_path):
    cache_file = str(tmp_path / 'cached')
    http._to_cache(cache_file, bytes)
    assert open(cache_file, 'rb').read() == exp_cached_bytes

def test_to_cache_cannot_create_cache_directory(mocker):
    mocker.patch('builtins.open')
    mkdir_mock = mocker.patch('upsies.utils.fs.mkdir', side_effect=OSError('No'))
    with pytest.raises(RuntimeError, match=r'^Unable to write cache file mock/path: No$'):
        http._to_cache('mock/path', b'data')
    assert mkdir_mock.call_args_list == [call('mock')]

def test_to_cache_cannot_write_cache_file(mocker):
    open_mock = mocker.patch('builtins.open')
    mkdir_mock = mocker.patch('upsies.utils.fs.mkdir')
    filehandle = open_mock.return_value.__enter__.return_value
    filehandle.write.side_effect = OSError('No')
    with pytest.raises(RuntimeError, match=r'^Unable to write cache file mock/path: No$'):
        http._to_cache('mock/path', b'data')
    assert mkdir_mock.call_args_list == [call('mock')]

def test_to_cache_writes_cache_file(mocker):
    open_mock = mocker.patch('builtins.open')
    mkdir_mock = mocker.patch('upsies.utils.fs.mkdir')
    filehandle = open_mock.return_value.__enter__.return_value
    assert http._to_cache('mock/path', b'data') is None
    assert open_mock.call_args_list == [call('mock/path', 'wb')]
    assert filehandle.write.call_args_list == [call(b'data')]
    assert mkdir_mock.call_args_list == [call('mock')]


USER_AGENT_SOURCE_URLS_MOCKED = [
    'http://get-me-some-user-agents.com',
    'http://info.org/user-agents',
    'http://thereyougo.net/have-some-user-agents.txt',
]

@pytest.mark.parametrize(
    argnames='cached_user_agent, downloaded_user_agents, exp_return_value, exp_mock_calls',
    argvalues=(
        # Get user agent from cache
        (
            'Cached (x86)', ['Downloaded (x86)'], 'Cached (x86)',
            [
                call._get_user_agent_from_cache(),
            ],
        ),
        # Get user agent from first source URL
        (
            None, ['Downloaded (x86)', None, None], 'Downloaded (x86)',
            [call._get_user_agent_from_cache()]
            + [
                call._get_user_agent_from_url(url)
                for url in USER_AGENT_SOURCE_URLS_MOCKED
            ][:1]
            + [call._cache_user_agent('Downloaded (x86)')],
        ),
        # Get user agent from second source URL
        (
            None, [None, 'Downloaded (x86)', None], 'Downloaded (x86)',
            [call._get_user_agent_from_cache()]
            + [
                call._get_user_agent_from_url(url)
                for url in USER_AGENT_SOURCE_URLS_MOCKED
            ][:2]
            + [call._cache_user_agent('Downloaded (x86)')],
        ),
        # Get user agent from DEFAULT_USER_AGENT
        (
            None, [None] * len(USER_AGENT_SOURCE_URLS_MOCKED), http.DEFAULT_USER_AGENT,
            [call._get_user_agent_from_cache()]
            + [
                call._get_user_agent_from_url(url)
                for url in USER_AGENT_SOURCE_URLS_MOCKED
            ],
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_get_popular_user_agent(cached_user_agent, downloaded_user_agents,
                                      exp_return_value, exp_mock_calls,
                                      mocker):
    mocker.patch('upsies.utils.http.USER_AGENT_SOURCE_URLS', USER_AGENT_SOURCE_URLS_MOCKED)
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch('upsies.utils.http._get_user_agent_from_cache', return_value=cached_user_agent),
        '_get_user_agent_from_cache',
    )
    mocks.attach_mock(
        mocker.patch('upsies.utils.http._get_user_agent_from_url', side_effect=downloaded_user_agents),
        '_get_user_agent_from_url',
    )
    mocks.attach_mock(
        mocker.patch('upsies.utils.http._cache_user_agent'),
        '_cache_user_agent',
    )
    return_value = await http.get_popular_user_agent()
    assert return_value == exp_return_value
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames='cached_user_agent, mtime_adjustment, mode, exp_return_value',
    argvalues=(
        (None, None, None, None),
        ('Cached (x86)', 0, 0o400, 'Cached (x86)'),
        ('Cached (x86)', http.USER_AGENT_CACHE_MAX_AGE - 1, 0o400, 'Cached (x86)'),
        ('Cached (x86)', http.USER_AGENT_CACHE_MAX_AGE + 1, 0o400, None),
        ('Cached (x86)', 0, 0o300, None),
    ),
)
def test_get_user_agent_from_cache(cached_user_agent, mtime_adjustment, mode, exp_return_value, mocker, tmp_path):
    if cached_user_agent is not None:
        cache_filepath = tmp_path / 'user-agent'
        cache_filepath.write_text(cached_user_agent)
        cache_filepath.chmod(mode)
        mtime = int(time.time()) - mtime_adjustment
        os.utime(cache_filepath, (time.time(), mtime))
        mocker.patch('upsies.utils.http._get_user_agent_cache_filepath', return_value=cache_filepath)

    return_value = http._get_user_agent_from_cache()
    assert return_value == exp_return_value


@pytest.mark.parametrize(
    argnames='response, acceptable_return_values',
    argvalues=(
        (
            (
                '<td> Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:123.0) like Gecko \n  </td>'
                '<td> Mozilla/5.0 (Linux; Android 13) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.5359.128 Mobile Safari/537.36 \n  </td>'
                '<td> Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:123.0) Gecko/20100101 Firefox/123.0 \n  </td>'
                '<td> Mozilla/5.0 (Macintosh; Intel Mac OS X 13.1; rv:123.0) Gecko/20100101 Firefox/123.0 \n  </td>'
                '<td> Mozilla/5.0 (X11; Linux i686; rv:123.0) Gecko/20100101 Firefox/123.0 \n  </td>'
                '<td> Mozilla/5.0 (X11; Linux x86_64; rv:123.0) Gecko/20100101 Firefox/123.0 \n  </td>'
                '<td> Mozilla/5.0 (X11; Ubuntu; Linux i686; rv:123.0) Gecko/20100101 Firefox/123.0  \n </td>'
                '<td> Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:123.0) Gecko/20100101 Firefox/123.0  \n </td>'
                '<td> Mozilla/5.0 (X11; Fedora; Linux x86_64; rv:123.0) Gecko/20100101 Firefox/123.0  \n </td>'
            ),
            [
                'Mozilla/5.0 (X11; Linux i686; rv:123.0) Gecko/20100101 Firefox/123.0',
                'Mozilla/5.0 (X11; Linux x86_64; rv:123.0) Gecko/20100101 Firefox/123.0',
                'Mozilla/5.0 (X11; Ubuntu; Linux i686; rv:123.0) Gecko/20100101 Firefox/123.0',
                'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:123.0) Gecko/20100101 Firefox/123.0',
                'Mozilla/5.0 (X11; Fedora; Linux x86_64; rv:123.0) Gecko/20100101 Firefox/123.0',
            ],
        ),
        (
            errors.RequestError('404 or something'),
            [None],
        ),
        (
            '<td>no user agents to be found here</td>',
            [None],
        ),
    ),
)
@pytest.mark.asyncio
async def test_get_user_agent_from_url(response, acceptable_return_values, mocker, tmp_path):
    if isinstance(response, Exception):
        get_mock = mocker.patch('upsies.utils.http.get', side_effect=response)
    else:
        get_mock = mocker.patch('upsies.utils.http.get', return_value=response)

    url = 'http://list-of-user-agents.com'

    return_value = await http._get_user_agent_from_url(url)
    assert return_value in acceptable_return_values

    assert get_mock.call_args_list == [call(url, cache=False, user_agent=False)]


@pytest.mark.parametrize(
    argnames='user_agent, success, exp_cached_user_agent',
    argvalues=(
        ('Fresh (x86)', True, 'Fresh (x86)'),
        ('Fresh (x86)', False, None),
    ),
)
def test_cache_user_agent(user_agent, success, exp_cached_user_agent, mocker, tmp_path):
    cache_filepath = tmp_path / 'user-agent'
    try:
        if not success:
            cache_filepath.parent.chmod(0o000)
        _get_user_agent_cache_filepath_mock = mocker.patch(
            'upsies.utils.http._get_user_agent_cache_filepath',
            return_value=cache_filepath,
        )

        http._cache_user_agent(user_agent)
    finally:
        cache_filepath.parent.chmod(0o700)

    assert _get_user_agent_cache_filepath_mock.call_args_list == [call()]
    if success:
        assert cache_filepath.exists()
        assert cache_filepath.read_text() == user_agent
    else:
        assert not cache_filepath.exists()

def test_get_user_agent_cache_filepth(mocker):
    cache_directory = '/path/to/cache'
    _get_cache_directory_mock = mocker.patch('upsies.utils.http._get_cache_directory', return_value=cache_directory)
    return_value = http._get_user_agent_cache_filepath()
    assert return_value == os.path.join(cache_directory, 'user-agent')
    assert _get_cache_directory_mock.call_args_list == [call()]



"""
Find, download and re-upload poster for movie, series or season
"""

import os

from .. import errors, utils
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PosterJob(JobBase):
    """
    Find poster image on a :mod:`webdb <upsies.utils.webdbs>`, download it and
    optionally upload it to :mod:`image host <upsies.utils.imghosts>`

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``finding_poster``
            Emitted before `webdb` is queried for a poster. Registered
            callbacks get the ID of the release.

        ``found_poster``
            Emitted after the poster was found. Registered callbacks get the URL
            to the poster image as a positional argument.

        ``downloading_poster``
            Emitted before the poster is downloaded. Registered callbacks get
            the URL to the poster image as a positional argument.

        ``downloaded_poster``
            Emitted after the poster was downloaded. Registered callbacks get
            the file path to the downloaded poster image as a positional
            argument.

        ``resizing_poster``
            Emitted before the poster is resized. Registered callbacks get the
            file path to the original poster image as a positional argument.

        ``resized_poster``
            Emitted after the poster was resized. Registered callbacks get the
            file path to the resized poster image as a positional argument.

        ``uploading_poster``
            Emitted before the poster is uploaded. Registered callbacks get the
            file path to the resized poster image as a positional argument.

        ``uploaded_poster``
            Emitted after the poster is uploaded. Registered callbacks get the
            URL to the resized poster image as a positional argument.
    """

    name = 'poster'
    label = 'Poster'

    cache_id = None
    """
    This job doesn't cache its output

    If the created poster is deleted, its cached output path is still
    regurgitated even though it doesn't exist anymore.

    Everything is already cached by the utilities, except for the image
    resizing, which should be reasonably fast.
    """

    def initialize(self, *, webdb, webdb_id, season=None, imghost=None, width=512, height=1024):
        """
        Set internal state

        :param webdb: Instance of :class:`~.WebDbApiBase` subclass
        :param webdb_id: ID for `webdb`
        :param season: Season number or `None`
        :param imghost: Instance of :class:`~.ImageHostBase` subclass or `None`
        :param int width: Maximum poster width in pixels
        :param int height: Maximum poster height in pixels

        `webdb_id` and/or `webdb` can also be `None`. In that case, this job
        must be created with `autostart` set to False. `webdb_id` and/or `webdb`
        must then be passed as keyword arguments to :meth:`start`.
        """
        self._webdb = webdb
        self._webdb_id = webdb_id
        self._season = season
        self._imghost = imghost
        self._width = int(width)
        self._height = int(height)

        self.signal.add('finding_poster')
        self.signal.add('found_poster')
        self.signal.add('downloading_poster')
        self.signal.add('downloaded_poster')
        self.signal.add('resizing_poster')
        self.signal.add('resized_poster')
        self.signal.add('uploading_poster')
        self.signal.add('uploaded_poster')

    def start(self, *, webdb=None, webdb_id=None, season=None):
        """
        Same as :meth:`~.JobBase.start`, but with the added possibility of providing
        `webdb`, `webdb_id` and `season` after initialization
        """
        # Set internal states
        super().start()

        _log.debug('Starting poster job: %r, %r, %r', webdb, webdb_id, season)
        self.attach_task(
            coro=self._work(
                webdb=webdb or self._webdb,
                webdb_id=webdb_id or self._webdb_id,
                season=season or self._season,
            ),
            finish_when_done=True,
        )

    async def _work(self, webdb, webdb_id, season=None):
        original_url = await self._get_poster_url(webdb, webdb_id, season)
        if original_url:
            original_filepath = self._get_poster_filepath(webdb, webdb_id, season, original_url)
            original_filepath = await self._download_poster(original_url, original_filepath)
            if original_filepath:
                resized_filepath = self._resize_poster(original_filepath)
                if resized_filepath:
                    if self._imghost:
                        resized_url = await self._upload_poster(resized_filepath)
                        self.send(resized_url)
                    else:
                        self.send(resized_filepath)

    async def _get_poster_url(self, webdb, webdb_id, season):
        self.signal.emit('finding_poster', webdb_id)
        try:
            url = await webdb.poster_url(webdb_id, season=season)
        except errors.RequestError as e:
            self.error(f'Failed to find poster: {e}')
        else:
            if not url:
                self.error('Failed to find poster: There is no poster')
            else:
                _log.debug('Found poster for %s: %r', webdb_id, url)
                self.signal.emit('found_poster', url)
                return url

    def _get_poster_filepath(self, webdb, webdb_id, season, url):
        filename = f'{webdb.name}={utils.fs.sanitize_filename(webdb_id)}'
        if season:
            season = str(utils.release.Episodes.from_string(f'S{season}'))
            filename += f'.{season}'

        url_extension = utils.fs.file_extension(url)
        if url_extension:
            filename += f'.{url_extension}'

        return os.path.join(self.cache_directory, filename)

    async def _download_poster(self, url, filepath):
        self.signal.emit('downloading_poster', url)
        try:
            filepath_sanitized = await utils.http.download(
                url,
                filepath,
                cache=not self.ignore_cache,
            )
        except errors.RequestError as e:
            self.error(f'Failed to download poster: {e}')
        else:
            self.signal.emit('downloaded_poster', filepath_sanitized)
            return filepath_sanitized

    def _resize_poster(self, filepath):
        filename = '.'.join((
            utils.fs.basename(utils.fs.strip_extension(filepath)),
            f'{self._width}x{self._height}',
            utils.fs.file_extension(filepath),
        ))

        self.signal.emit('resizing_poster', filepath)
        try:
            resized_filepath = utils.image.resize(
                filepath,
                target_directory=self.home_directory,
                target_filename=filename,
                width=self._width,
                height=self._height,
            )
        except errors.ImageResizeError as e:
            self.error(f'Failed to resize poster: {e}')
        else:
            self.signal.emit('resized_poster', resized_filepath)
            return resized_filepath

    async def _upload_poster(self, filepath):
        self.signal.emit('uploading_poster', filepath)
        try:
            resized_url = await self._imghost.upload(filepath)
        except errors.RequestError as e:
            self.error(f'Failed to upload poster: {e}')
        else:
            self.signal.emit('uploaded_poster', resized_url)
            return resized_url
